"""
Support for working with indices of numeric arrays.
"""

import contextlib
import numbers
import typing

import numpy
import numpy.typing

from .. import container
from .. import exceptions
from .. import measurable
from ._data import isintegral
from ._mixins import Operators
from . import _implementation
from . import _operators
from . import _objects
from ._operations import OperandTypeError


T = typing.TypeVar('T')

I = typing.TypeVar('I', bound=typing.SupportsInt)


class Object(_objects.Object[T], Operators):
    """Base class for index objects."""

    __eq__ = _implementation.binary('eq')
    __ne__ = _implementation.binary('ne')
    __lt__ = _implementation.binary('lt')
    __gt__ = _implementation.binary('gt')
    __le__ = _implementation.binary('le')
    __ge__ = _implementation.binary('ge')

    __abs__ = _implementation.unary('abs')
    __pos__ = _implementation.unary('pos')
    __neg__ = _implementation.unary('neg')

    __add__, __radd__ = _implementation.doublet('add')
    __sub__, __rsub__ = _implementation.doublet('sub')
    __mul__, __rmul__ = _implementation.doublet('mul')
    __floordiv__, __rfloordiv__ = _implementation.doublet('floordiv')
    __mod__, __rmod__ = _implementation.doublet('mod')
    __pow__, __rpow__ = _implementation.doublet('pow')

    def _get_numpy_array(self):
        return numpy.array(self.data)

    _OPERATORS = {}
    _FUNCTIONS = {}


class Value(Object[int]):
    """An integral index value."""

    def __index__(self) -> int:
        """Called for use as an index or to convert to integer type."""
        return self.data

    def shift(self, __x: I, floor: I=None, ceil: I=None):
        """Shift index by a constant value."""
        index = int(self.data) + int(__x)
        if floor is not None and index < floor:
            return Value(floor)
        if ceil is not None and index > ceil:
            return Value(ceil)
        return Value(index)


ObjectType = typing.TypeVar('ObjectType', bound=Object)
"""Type variable bound to `~index.Object`."""


class Sequence(Object[numpy.typing.NDArray[numpy.integer]]):
    """A sequence of integral index values."""

    def __str__(self) -> str:
        if len(self.data) <= 4:
            return ', '.join(str(i) for i in self.data)
        indices = [*self.data[:2], '...', *self.data[-2:]]
        return ', '.join(str(i) for i in indices)

    # XXX: Should instances be iterators rather than iterables? See
    # https://stackoverflow.com/a/24377 for more information and suggestions.
    def __iter__(self):
        """Called for iter(self)."""
        yield from self.data

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.data)

    def __getitem__(self, __i: typing.SupportsIndex):
        """Called for self[i]."""
        data = self.data[__i]
        try:
            len(data)
        except TypeError:
            result = Value.spawn(data)
        else:
            result = self.spawn(data)
        return result

    def shift(self, __x: I, floor: I=None, ceil: I=None):
        """Shift indices by a constant value."""
        indices = numpy.array(self.data) + int(__x)
        if floor is not None:
            indices[indices < floor] = floor
        if ceil is not None:
            indices[indices > ceil] = ceil
        return Sequence(indices)


IndexType = typing.TypeVar('IndexType', Value, Sequence)

IndexLike = typing.Union[Value, Sequence]


def implements(operation):
    """Register `operation` as the implementation for all objects."""
    def wrapped(f):
        Value.implements(operation)(f)
        Sequence.implements(operation)(f)
        return f
    return wrapped


@implements('eq')
@exceptions.convert(OperandTypeError, False)
def eq(a, b, /):
    """Compute a == b."""
    return comparative(_operators.eq, a, b)


@implements('ne')
@exceptions.convert(OperandTypeError, True)
def ne(a, b, /):
    """Compute a != b."""
    return comparative(_operators.ne, a, b)


@implements('abs')
def abs(a):
    """Compute abs(a)."""
    return unary(_operators.abs, a)


@implements('pos')
def pos(a):
    """Compute +a."""
    return unary(_operators.pos, a)


@implements('neg')
def neg(a):
    """Compute -a."""
    return unary(_operators.neg, a)


@implements('round')
def round(a, /, ndigits: int=None):
    """Compute round(a[, ndigits])."""
    return unary(_operators.round, a, ndigits=ndigits)


@implements('lt')
@exceptions.convert(OperandTypeError, NotImplemented)
def lt(a, b):
    """Compute a < b."""
    return comparative(_operators.lt, a, b)


@implements('le')
@exceptions.convert(OperandTypeError, NotImplemented)
def le(a, b):
    """Compute a <= b."""
    return comparative(_operators.le, a, b)


@implements('gt')
@exceptions.convert(OperandTypeError, NotImplemented)
def gt(a, b):
    """Compute a > b."""
    return comparative(_operators.gt, a, b)


@implements('ge')
@exceptions.convert(OperandTypeError, NotImplemented)
def ge(a, b):
    """Compute a >= b."""
    return comparative(_operators.ge, a, b)


@implements('add')
def add(a, b):
    """Compute a + b."""
    return arithmetic(_operators.add, a, b)


@implements('sub')
def sub(a, b):
    """Compute a - b."""
    return arithmetic(_operators.sub, a, b)


@implements('mul')
def mul(a, b):
    """Compute a * b."""
    return arithmetic(_operators.mul, a, b)


@implements('floordiv')
def floordiv(a, b):
    """Compute a // b."""
    return arithmetic(_operators.floordiv, a, b)


@implements('mod')
def mod(a, b):
    """Compute a % b."""
    return arithmetic(_operators.mod, a, b)


@implements('pow')
def pow(a, b, mod: int=None):
    """Compute a ** b."""
    negative_exponent = False
    try:
        negative_exponent = b < 0
    except ValueError:
        negative_exponent = all(b < 0)
    if negative_exponent:
        raise ValueError(
            "Cannot raise index to a negative integral power"
        ) from None
    return arithmetic(_operators.pow, a, b, mod=mod)


A = typing.TypeVar('A', bound=numbers.Real)
R = typing.TypeVar('R', bound=numbers.Real)

def unary(
    f: typing.Callable[[A], R],
    a: IndexType,
    **kwargs
) -> IndexType:
    """Implement a unary operation on `Value` or `Sequence`."""
    if isinstance(a, Sequence):
        return a.spawn([f(v, **kwargs) for v in a])
    if isinstance(a, Value):
        return a.spawn(f(a.data, **kwargs))
    raise OperandTypeError(a)


B = typing.TypeVar('B', bound=numbers.Real)

BoolLike = typing.Union[bool, typing.Sequence[bool]]

def comparative(
    f: typing.Callable[[A, B], BoolLike],
    a: IndexType,
    b: IndexType,
) -> BoolLike:
    """Implement a binary comparative operation on `Value` or `Sequence`."""
    r = binary(f, a, b)
    if isinstance(a, Value) and isinstance(b, Value):
        return bool(r)
    return numpy.array(r, dtype=bool)


@typing.overload
def arithmetic(
    f: typing.Callable[[A, B], R],
    a: IndexType,
    b: Sequence,
    **kwargs
) -> Sequence: ...

@typing.overload
def arithmetic(
    f: typing.Callable[[A, B], R],
    a: Sequence,
    b: IndexType,
    **kwargs
) -> Sequence: ...

@typing.overload
def arithmetic(
    f: typing.Callable[[A, B], R],
    a: Value,
    b: Value,
    **kwargs
) -> Value: ...

@typing.overload
def arithmetic(
    f: typing.Callable[[A, B], R],
    a: typing.Union[IndexType, numbers.Real],
    b: typing.Union[IndexType, numbers.Real],
    **kwargs
) -> IndexType: ...

def arithmetic(f, a, b, **kwargs):
    """Implement a binary arithmetic operation on `Value` or `Sequence`."""
    r = binary(f, a, b, **kwargs)
    if isinstance(a, (Value, Sequence)) and isinstance(b, (Value, Sequence)):
        if all(isinstance(i, Value) for i in (a, b)):
            return _new_or_result(Value.spawn, r)
        return _new_or_result(Sequence.spawn, r)
    if isinstance(a, (Value, Sequence)) and isinstance(b, numbers.Integral):
        return _new_or_result(a.spawn, r)
    if isinstance(a, numbers.Integral) and isinstance(b, (Value, Sequence)):
        return _new_or_result(b.spawn, r)
    raise OperandTypeError(a, b)


def binary(f, a, b, **kwargs):
    """Compute f(a, b, **kwargs) on numeric data."""
    if isinstance(a, Sequence) and isinstance(b, Sequence):
        return [f(x, y, **kwargs) for (x, y) in zip(a.data, b.data)]
    if isinstance(a, Sequence):
        if isinstance(b, Value):
            return [f(x, b.data, **kwargs) for x in a.data]
        if isintegral(b):
            return [f(x, b, **kwargs) for x in a.data]
        raise OperandTypeError(a, b)
    if isinstance(b, Sequence):
        if isinstance(a, Value):
            return [f(a.data, y, **kwargs) for y in b.data]
        if isintegral(a):
            return [f(a, y, **kwargs) for y in b.data]
        raise OperandTypeError(a, b)
    if isinstance(a, Value) and isinstance(b, Value):
        return f(a.data, b.data, **kwargs)
    if isinstance(a, Value) and isintegral(b):
        return f(a.data, b, **kwargs)
    if isintegral(a) and isinstance(b, Value):
        return f(a, b.data, **kwargs)
    raise OperandTypeError(
        f"Expected operands to be instances of {Value} or {Sequence}"
    ) from None


def _new_or_result(factory, x):
    """Helper for computing correct return type.
    
    If applying `factory` to `x` succeeds, this function will return the result.
    Otherwise, it will return `x`.
    """
    try:
        result = factory(x)
    except Exception:
        result = x
    return result


@typing.overload
def value(
    x: typing.Union[
        typing.SupportsInt,
        typing.Iterable[typing.SupportsInt],
        numpy.typing.NDArray[numpy.integer],
    ],
    /,
) -> Value:
    """Create an index value from a numeric object.

    Parameters
    ----------
    x
        An object that supports conversion to `int`, an iterable object with a
        single member that supports conversion to `int`, or a one-element
        `numpy` array with integral data type.

    Returns
    -------
    `~Value`

    Raises
    ------
    `TypeError`
        The argument contains more than one element.
    """

@typing.overload
def value(x: measurable.Type, /) -> Value:
    """Create an index value from a unitless measurable object.

    Parameters
    ----------
    x
        An object with a `data` attribute and a unit equal to `'1'`.

    Returns
    -------
    `~Value`

    Raises
    -----
    `TypeError`
        The given object is not unitless.
    """

@Value.register.factory
def value(x, /):
    """Create a single numeric index value."""
    parsed = value_args(x)
    if isinstance(parsed, typing.SupportsInt):
        if isinstance(parsed, numpy.ndarray) and parsed.size == 1:
            return Value(int(parsed[0]))
        return Value(int(parsed))
    raise TypeError(
        f"Cannot convert {x} to a single integer"
    ) from None


@typing.overload
def sequence(
    x: typing.Union[
        typing.SupportsInt,
        typing.Iterable[typing.SupportsInt],
        numpy.typing.NDArray[numpy.integer],
    ],
    /,
) -> Sequence:
    """Create an index sequence from an iterable object with integer-like values.

    Parameters
    ----------
    x
        An object that supports conversion to `int`, an iterable object with
        members that all support conversion to `int`, or a `numpy` array with
        integral data type.

    Returns
    -------
    `~Value`

    Raises
    ------
    `TypeError`
        The given object cannot be coerced into a one-dimensional array.
    """

@typing.overload
def sequence(x: measurable.Type, /) -> Sequence:
    """Create an index sequence from a unitless measurable object.

    Parameters
    ----------
    x
        An object with a `data` attribute and a unit equal to `'1'`.

    Returns
    -------
    `~Value`

    Raises
    -----
    `TypeError`
        The given object is not unitless.
    """

@Sequence.register.factory
def sequence(x, /):
    """Create a sequence of numeric index values."""
    parsed = sequence_args(x)
    if all(isinstance(i, typing.SupportsInt) for i in parsed):
        return Sequence(numpy.array(parsed, dtype=int))
    raise TypeError(
        f"The numerical values of {x} do not all support"
        " conversion to integral type"
    ) from None


def value_args(x, /):
    """Parse arguments to initialize `~Value`."""
    index = _value_args(x)
    if index is not None:
        return index
    indices = _sequence_args(x)
    if indices is not None:
        return indices[0]
    raise TypeError(
        f"Cannot initialize {Value} from {x}, which has type {type(x)}"
    ) from None


def sequence_args(x, /):
    """Parse arguments to initialize `~Sequence`."""
    indices = _sequence_args(x)
    if indices is not None:
        return indices
    index = _value_args(x)
    if index is not None:
        return [index]
    raise TypeError(
        f"Cannot initialize {Sequence} from {x}, which has type {type(x)}"
    ) from None


def _value_args(x, /):
    """Helper for `index_args`."""
    if isinstance(x, numpy.ndarray) and x.dtype == int:
        if x.size == 1:
            v = x[0] if x.ndim == 1 else x
            return int(v)
        raise TypeError(
            f"Can only initialize {Value} from a"
            " zero- or one-dimensional array-like object"
        ) from None
    if isinstance(x, (numbers.Integral, numpy.integer)):
        return x
    if isinstance(x, (bytes, str)):
        try:
            index = int(x)
        except ValueError:
            return # Force a TypeError in factory
        return index
    if isinstance(x, measurable.Type):
        if x.isunitless:
            return x.data
        raise ValueError(
            f"Can only initialize {Value} from a unitless object"
        ) from None


def _sequence_args(x, /):
    """Helper for `indices_args`."""
    if isinstance(x, range):
        return x
    if isinstance(x, (tuple, list)):
        if all(isinstance(i, str) for i in x):
            with contextlib.suppress(ValueError):
                return [int(i) for i in x]
        if all(isinstance(i, (numbers.Integral, numpy.integer)) for i in x):
            return x
    if isinstance(x, numpy.ndarray) and x.dtype == int:
        if x.size == 1:
            v = x[0] if x.ndim == 1 else x
            return numpy.array(int(v), ndmin=1)
        y = x.squeeze()
        ndim = y.ndim
        if ndim == 1:
            return y
        raise TypeError(
            f"Cannot initialize {Sequence} from a {ndim}-D array"
        ) from None
    if isinstance(x, (numbers.Integral, numpy.integer)):
        return [x]
    if isinstance(x, Sequence):
        return x.data
    if isinstance(x, measurable.Type):
        if x.isunitless:
            return _sequence_args(x.data)
        raise ValueError(
            f"The argument to {Sequence} must be unitless."
        ) from None


_BUILTIN_TYPES = (int, slice, type(...), type(None))

_INDEX_DTYPES = (
    bool,
    numpy.int8,
    numpy.int16,
    numpy.int32,
    numpy.int64,
    numpy.uint8,
    numpy.uint16,
    numpy.uint32,
    numpy.uint64,
)

def resolve(shape: typing.Sequence[int], args):
    """Convert `args` into array indices based on `shape`.

    This function will first normalize `args` via `~normalize`. If the result
    contains an `Ellipsis` (i.e., `...`), this function will expand the
    `Ellipsis` via `~expand`. Otherwise, it will return the normalized indices.
    """
    normalized = normalize(shape, args)
    if _has_ellipsis(normalized):
        return expand(len(shape), normalized)
    return normalized


def _has_ellipsis(args):
    """Helper for `~resolve`."""
    try:
        iter(args)
    except TypeError:
        return False
    return any(arg is Ellipsis for arg in args)


def normalize(shape: typing.Sequence[int], args):
    """Compute appropriate array indices from `args`.

    If all indices in `args` have a standard form involving slices, an ellipsis,
    or integers, including 

    - `array[:]`
    - `array[...]`
    - `array[i, :]`
    - `array[:, j]`
    - `array[i, j]`

    (where i and j are integers) this function will immediately return them. If
    `args` is a `numpy.ndarray` with boolean type, this method will immediately
    return it. Otherwise, it will extract a sequence of indices that represents
    the original dimensions of the data.
    """
    if _normalized(args):
        return args
    if not isinstance(args, (tuple, list)):
        args = [args]
    expanded = expand_ellipsis(len(shape), *args)
    indices = []
    for i, arg in enumerate(expanded):
        if isinstance(arg, slice):
            indices.append(range(shape[i]))
        elif isintegral(arg):
            indices.append((arg,))
        else:
            indices.append(arg)
    return numpy.ix_(*list(indices))


def _normalized(args):
    """Helper for `~normalize`."""
    return (
        container.hastype(args, _BUILTIN_TYPES, tuple, strict=True)
        or
        isinstance(args, numpy.ndarray) and args.dtype in _INDEX_DTYPES
    )


def expand(ndim: int, indices):
    """Expand `indices` so that they will index `ndim` dimensions."""
    if isinstance(indices, (list, tuple)):
        return expand_ellipsis(ndim, *indices)
    if indices == slice(None):
        return expand_ellipsis(ndim, ...)
    if isinstance(indices, type(...)):
        return expand_ellipsis(ndim, indices)
    if ndim > 1:
        return expand_ellipsis(ndim, indices, ...)
    return (indices,)


def expand_ellipsis(ndim: int, *args: typing.SupportsIndex):
    """Expand an `Ellipsis` into one or more `slice` objects."""
    # If there are more arguments than dimensions, something is wrong. The one
    # exception is when the final argument is an Ellipsis. This allows upstream
    # code to programmatically use `...` to capture 0 or more remaining
    # dimensions. For example, `self.mfp[0, 0, ...]` should be equal to
    # `self.mfp[0, 0, :, :]` whereas `self.r[0, 0, ...]` should be equal to
    # `self.r[0, 0]`.
    nargs = len(args)
    if nargs > ndim and not isinstance(args[-1], type(...)):
        raise IndexError(
            f"Too many indices ({nargs}) for {ndim} dimensions"
        ) from None
    # Convert numpy arrays to lists to avoid ValueError in `count(...)`.
    norm = [
        list(arg) if isinstance(arg, numpy.ndarray) else arg
        for arg in args
    ]
    # Count the number of ellipses.
    count = norm.count(...)
    # If there is no ellipsis, there's nothing to do.
    if count == 0:
        return args
    # If there is more than one ellipsis, raise an exception.
    if count > 1:
        raise IndexError(
            f"Index arguments may only contain a single ellipsis ('...')"
        ) from None
    # Expand arguments into
    # 1) all arguments before the ellipsis
    # 2) a slice for every dimension represented by the ellipsis
    # 3) all arguments after the ellipsis
    nslice = ndim - nargs + 1
    ellpos = args.index(Ellipsis)
    return (
        *args[slice(0, ellpos)],
        *([slice(None)] * nslice),
        *args[slice(ellpos+1, nargs+1)],
    )


def restrict(
    array: numpy.ndarray,
    *targets: float,
    axis: typing.SupportsIndex,
) -> range:
    """Restrict indices of `array` around the target value(s) along `axis`.

    Parameters
    ----------
    array : array-like
        The array containing `target`.

    *targets : real
        The numerical value(s) in `array` whose index to bound. There must be at
        least one target value.

    axis : integral, default=-1
        The axis in `array` that contains the target value(s).

    Returns
    -------
    `range`

    Raises
    ------
    `ValueError`:
        The caller did not pass any target values.
    """
    if len(targets) == 0:
        raise ValueError("No target values to bound") from None
    if len(targets) == 1:
        lower, upper = _compute_bounds(array, targets[0], axis)
        return range(lower, upper+1)
    lower = None
    upper = None
    for target in targets:
        lo, hi = _compute_bounds(array, target, axis)
        lower = lo if lower is None else min(lower, lo)
        upper = hi if upper is None else max(upper, hi)
    return range(lower, upper+1)


def _compute_bounds(
    array: numpy.ndarray,
    target: float,
    axis: typing.SupportsIndex,
) -> typing.Tuple[int, int]:
    """Compute the indices in `array` that bound `target` along `axis`.

    This function is a helper for `~restrict`.
    """
    bounds = _get_index_bounds(array, target, axis=axis)
    if bounds.ndim == 1:
        size = bounds.size
        if size != 2:
            raise ValueError(
                f"Attempt to bound {target} along axis {axis} in {array}"
                f" produced a {size}-element array."
            ) from None
        return tuple(bounds)
    if bounds.ndim == 2:
        lower = numpy.min(bounds[:, 0])
        upper = numpy.max(bounds[:, 1])
        return lower, upper
    # NOTE: We can get `lower` and `upper` for N-D arrays by replacing `:` with
    # `...`. The problem is that they will be (N-1)-D arrays that we will then
    # need to split into individual index vectors before returning.
    raise NotImplementedError(
        f"Cannot compute bounds of {bounds.ndim}-D array"
    ) from None


def _get_index_bounds(
    array: numpy.typing.ArrayLike,
    target: numbers.Real,
    axis: typing.SupportsIndex=None,
) -> numpy.typing.NDArray[numpy.integer]:
    """Find the indices bounding the target value.

    This function is a helper for `~restrict`. It returns an array containing
    the indices bounding `target` at each slice along `axis`. The shape will be
    the same as that of `array` except for `axis`, which will be 2.
    """
    if axis is None:
        axis = -1
    return numpy.apply_along_axis(
        _find_1d_indices,
        int(axis),
        numpy.asarray(array, dtype=numpy.float64),
        float(target),
    )


def _find_1d_indices(
    array: numpy.ndarray,
    target: float,
) -> typing.Tuple[int, int]:
    """Find the bounding indices in a 1-D array."""
    leq = array <= target
    lower = numpy.where(leq)[0].max() if any(leq) else 0
    geq = array >= target
    upper = numpy.where(geq)[0].min() if any(geq) else len(array)-1
    return lower, upper





"""
Utilities for the `eprem` package.
"""

import abc
import functools
import inspect
import numbers
import textwrap
import typing

import numpy

# NOTE: This module should not import any modules from the `eprem` package.


T = typing.TypeVar('T')


def strargs(*pos, **kwd) -> str:
    """Format arguments as if explicitly passed."""
    def strfmt(x):
        return repr(x) if isinstance(x, str) else str(x)
    strpos = ', '.join(strfmt(x) for x in pos)
    if not kwd:
        return strpos
    strkwd = ', '.join(f"{k}={strfmt(v)}" for k, v in kwd.items())
    if not pos:
        return strkwd
    return f"{strpos}, {strkwd}"


def underline(text: str):
    """Print underlined text."""
    dashes = '-' * len(text)
    print(f"\n{text}")
    print(dashes)


def doc2help(
    obj: typing.Union[str, typing.Any], /,
    mode: str='summary',
    replacements: typing.Mapping[str, str]=None,
) -> str:
    """Format the target docstring for use as help text.
    
    Parameters
    ----------
    obj
        The target object. If `obj` is a string, this function will directly
        operate on `obj`. If `obj` is an object with a `__doc__` attribute, this
        function will operate on the docstring contained in `obj.__doc__`.

    mode : str, default='sumamry'
        Controls the formatting of the result. In 'summary' mode (the default),
        this function will return the initial (i.e., summary) line of the target
        docstring as written. In 'phrase' mode, this function will return the
        initial line of the docstring in lower case, without a trailing period.
        In 'full' mode, this function will format and return the entire
        docstring.

    replacements : mapping from string to string
        Key-value pairs of strings to replace in the target docstring, where the
        key is the current string and the value is the new string (as in
        ``str.replace``).
    """
    try:
        target = obj if isinstance(obj, str) else str(obj.__doc__)
    except AttributeError:
        raise TypeError(f"Cannot create help text from {obj!r}") from None
    for k, v in (replacements or {}).items():
        target = target.replace(k, v)
    doclines = target.lstrip('\n').split('\n')
    summary = doclines[0]
    if mode == 'full':
        wrapped = textwrap.wrap(
            ' '.join(s.lstrip(' \t') for s in doclines[1:]),
            width=70
        )
        body = '\n'.join(wrapped).lstrip(' ')
        return f"{summary}\n\n{body}"
    if mode == 'phrase':
        return summary[0].lower() + summary[1:].rstrip('.')
    return summary


class InstanceSet(abc.ABCMeta):
    """A metaclass for sets of singletons.

    Using this class as a metaclass will ensure that only one instance of the
    object class exists for a given combination of initializing arguments. Given
    a concrete implementation of `_generate_key`, this metaclass will simply
    return the appropriate existing instance instead of creating a new one.

    See https://refactoring.guru/design-patterns/singleton/python/example
    """

    _instances = {}

    def __call__(self, *args, **kwargs):
        """Ensure that only one instance of the given object exists."""
        key = self._generate_key(*args, **kwargs)
        if key not in self._instances:
            self._instances[key] = super().__call__(*args, **kwargs)
        return self._instances[key]

    @abc.abstractmethod
    def _generate_key(self, *args, **kwargs):
        """Generate a unique instance key.
        
        Concrete implementations of this method must map the given arguments to
        a valid dictionary key, which will be associated with a unique instance.
        """
        raise TypeError(
            "Can't generate unique mapping key"
            f" from {strargs(*args, **kwargs) or 'arguments'}"
        ) from None


def apply(
    methods: typing.Iterable[typing.Callable[..., T]],
    *args,
    **kwargs,
) -> typing.Optional[T]:
    """Apply the given methods until one returns a non-null result."""
    gen = (method(*args, **kwargs) for method in methods)
    if result := next((match for match in gen if match), None):
        return result


def str2repr(__cls: typing.Type[T]):
    """Decorator that provides ``T.__repr__`` based on ``T.__str__``.
    
    Examples
    --------
    The following will define ``Class.__repr__``::

        @str2repr
        class Class:
            def __init__(self, v) -> None:
                self.v = v
            def __str__(self) -> str:
                return f"v={self.v}"

    ```
    >>> c = Class(3)
    >>> print(c)
    3
    >>> c
    Class(v=3)
    ```
    """
    def repr(self: T):
        c = self.__class__
        name = c.__name__
        if '__str__' not in c.__dict__:
            return name
        return f"{name}({self})"
    __cls.__repr__ = repr
    return __cls


def autostr(_obj: typing.Type[T]=None, style: str='default', **kwargs):
    """Decorator that generates `T.__repr__` and `T.__str__` if necessary.
    
    Examples
    --------
    The following will define `Class.__repr__` from `Class.__str__`::

        @autostr
        class Class:
            def __init__(self, v) -> None:
                self.v = v
            def __str__(self) -> str:
                return f"v={self.v}"

    >>> c = Class(3)
    >>> print(c)
    3
    >>> c
    Class(v=3)

    Decorating a class that already defines `__str__` and `__repr__` will not
    modify those methods::

        @autostr
        class Brak:
            def __str__(self) -> str:
                return "Don't touch me."
            def __repr__(self) -> str:
                return "Hey! Don't touch me!"

    >>> b = Brak()
    >>> print(b)
    Don't touch me.
    >>> b
    Hey! Don't touch me!
    """
    @functools.wraps(_obj)
    def wrapper(obj):
        if style == 'numpy':
            return _autostr_numpy(obj, **kwargs)
        if style == 'default':
            return _autostr_default(obj)
    if _obj is None:
        return wrapper
    return wrapper(_obj)


def _autostr_numpy(obj, **kwargs):
    """Helper for `~etc.autostr` in 'numpy' mode."""
    options = {
        'threshold': kwargs.pop('threshold', 4),
        'edgeitems': kwargs.pop('edgeitems', 2),
        'separator': kwargs.pop('separator', ', '),
        'precision': kwargs.pop('precision', 3),
        'floatmode': kwargs.pop('floatmode', 'maxprec_equal'),
    }
    options.update(kwargs)
    def _str(self: T) -> str:
        return numpy.array2string(numpy.array(self), **options)
    prefix = kwargs.pop('prefix', f"{obj.__name__}(")
    suffix = kwargs.pop('suffix', ")")
    def _repr(self: T) -> str:
        content = numpy.array2string(
            numpy.array(self),
            prefix=prefix,
            suffix=suffix,
            **options
        )
        return f"{prefix}{content}{suffix}"
    if '__str__' not in obj.__dict__:
        obj.__str__ = _str
    if '__repr__' not in obj.__dict__:
        obj.__repr__ = _repr
    return obj


def _autostr_default(obj):
    """Helper for `~etc.autostr` in 'default' mode."""
    def _repr(self: T) -> str:
        """An unambiguous representation of this object."""
        c = self.__class__
        name = c.__name__
        if not _defines_str(c):
            return name
        return f"{name}({self})"
    if '__repr__' not in obj.__dict__:
        obj.__repr__ = _repr
    return obj


def _defines_str(this: type):
    """True if `this` type defines a ``__str__`` method.
    
    This function is designed to prevent recursion in `~etc._autostr_default`.
    It checks for ``__str__`` in the ``__dict__`` attribute of `this` type, or
    for a ``__str__`` attribute of `this` type that is not ``object.__str__``. A
    false result indicates to `~etc._autostr_default` that implicitly calling
    `this.__str__` from `_repr` will cause recursion.
    """
    return '__str__' in this.__dict__ or (
        hasattr(this, '__str__')
        and getattr(this, '__str__') != getattr(object, '__str__')
    )


_ST = typing.TypeVar('_ST', bound='Singleton')


class Singleton:
    """A simple base class for creating singletons."""

    _exists = False
    _instance = None

    def __new__(cls: typing.Type[_ST], *args, **kwargs) -> _ST:
        if cls._exists:
            return cls._instance
        new = super(Singleton, cls).__new__(cls)
        cls._exists = True
        cls._instance = new
        return new


class NothingType(Singleton):
    """An object that represents nothing in a variety of ways."""

    def __getitem__(self, index: typing.Any) -> None:
        """Return `None`, regardless of `index`."""
        return None

    def __call__(self, *args, **kwargs) -> None:
        """Return `None`, regardless of input."""
        return None

    def __len__(self) -> int:
        """This object has zero length."""
        return 0

    def __contains__(self, key: str) -> bool:
        """This object contains nothing."""
        return False

    def __iter__(self) -> typing.Iterable:
        """Return an empty iterator.
        
        Thanks to https://stackoverflow.com/a/36658865/4739101 for this one.
        """
        yield from ()

    def __next__(self):
        """There is always nothing left."""
        raise StopIteration

    def __bool__(self) -> bool:
        """This object is always false."""
        return False

    def __repr__(self) -> str:
        """This object knows that it is nothing."""
        return "Nothing"


Nothing = NothingType()
"""A unique object that represents nothing."""


class SentinelInitError(Exception):
    """Attempt to directly instantiate a new sentinel object."""


class SentinelValueError(ValueError):
    """Value error while creating a sentinel object."""


_sentinels = {}
"""Internal mapping of singleton key-value pairs (a.k.a sentinels)."""


def _sentinel_init(f):
    """Decorator for `~_SentinelType.__init__`.
    
    The core purpose of this decorator function is to force
    `~_SentinelType.__init__` to unconditionally raise `NotImplementedError`. It
    does so by consuming all user arguments, which would cause the runtime
    environment to raise a `ValueError` before `~_SentinelType.__init__` could
    raise `NotImplementedError`.
    """
    exception = SentinelInitError
    def _parse_args(*args, **kwargs):
        """Create a string to represent the user arguments.
        
        If this function detects that `*args` and `**kwargs` represent valid
        arguments to `~sentinel`, it will return a string representing the
        corresponding arguments. Otherwise, it will return `None`. In the former
        case, `_make_errmsg` will create a message that includes the appropriate
        function call. In the latter case, `_make_errmsg` will create a generic
        message that defers to the `~sentinel` documentation.
        """
        if len(args) == 2 and not kwargs:
            name = args[0]
            if isinstance(name, str):
                return f"{name!r}, {args[1]}"
        if len(args) == 1 and list(kwargs) == ['value']:
            name = args[0]
            if isinstance(name, str):
                return f"{name!r}, value={kwargs['value']}"
        if not args and list(kwargs) == ['name', 'value']:
            name = kwargs['name']
            if isinstance(name, str):
                return f"name={name!r}, value={kwargs['value']}"
    def _make_errmsg(*args, **kwargs):
        """Create an appropriate error message for the given arguments."""
        strfunc = f"{__name__}.sentinel"
        if strargs := _parse_args(*args, **kwargs):
            return f"Please call {strfunc}({strargs}) instead."
        return f"Please see {strfunc} for more information."
    expected = NotImplementedError
    def method(self, *args, **kwargs):
        """Wrapper method for `__init__`."""
        errmsg = _make_errmsg(*args, **kwargs)
        try:
            f(self)
        except expected as err:
            raise exception(
                f"Cannot directly instantiate {type(self)}. {errmsg}"
            ) from err
        raise RuntimeError(
            f"Failed to catch {expected}"
            f" during attempt to instantiate {type(self)}"
        ) from None
    return method


class _SentinelType(typing.Generic[T]):
    """Factory class for singleton key-value pairs."""

    _name: str
    _value: T

    @classmethod
    def _create(cls, name: str, value: T):
        """Pseudo-private instantiation method."""
        if value in _sentinels:
            return _sentinels[value]
        instance = cls.__new__(cls)
        instance._name = name
        instance._value = value
        _sentinels[name] = instance
        return instance

    @_sentinel_init
    def __init__(self) -> None:
        """It is not possible to instantiate this class.
        
        Please use the `~sentinel` factory function.
        """
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        """Called for self == other."""
        if not isinstance(other, _SentinelType):
            return self._value == other
        return self is other

    def __bool__(self):
        """Called for bool(self)."""
        if self._value in (None, False):
            return False
        return True

    def __hash__(self):
        """Called for hash(self)."""
        return hash(self._value)

    def __repr__(self) -> str:
        """The canonical name of this sentinel object."""
        return self._name


def sentinel(name: str, value: T) -> _SentinelType:
    """Create a new sentinel object.
    
    This function provides the public interface for creating new sentinel
    objects, which are essentially singleton key-value pairs.

    Parameters
    ----------
    name : string
        The name of the new sentinel object. This should be the same as the name
        of the variable to which it is bound (e.g., `X = sentinel('X', -1)`).
    value
        The value that the new object will represent.

    Returns
    -------
    `~_SentinelType`
    """
    if isinstance(value, _SentinelType):
        raise SentinelValueError(
            f"Sentinel object {__name__}.{value!r} already exists"
        ) from None
    return _SentinelType._create(name, value)


PASS = sentinel('PASS', True)
"""Sentinel object equivalent to `True`."""

NULL = sentinel('NULL', None)
"""Sentinel object equivalent to `None`."""

FAIL = sentinel('FAIL', False)
"""Sentinel object equivalent to `False`."""


def isnull(this: typing.Any) -> bool:
    """True if `this` is empty but not if it's 0.

    This function allows the calling code to programmatically test for objects
    that are logically ``False`` except for numbers equivalent to 0.
    """
    if isinstance(this, numbers.Number):
        return False
    size = getattr(this, 'size', None)
    if size is not None:
        return size == 0
    try:
        result = not bool(this)
    except ValueError:
        result = all((isnull(i) for i in this))
    return result


def join(x: typing.Iterable[str], c: str='and', /, quoted: bool=False):
    """Join strings, with a conjunction before the final item.
    
    Parameters
    ----------
    x : iterable of strings
        The strings to join. This function will convert `x` into a `list`.
    
    c : string
        The conjunction to insert before the final item, if `x` contains more
        than one string.

    quoted : bool, default=False
        If true, quote each string in `x`.

    Notes
    -----
    - This function will insert the conjunction as given. It is the user's
      responsibility to pass an appropriate argument.
    - This function implements the `quoted` option by calling `repr` on each
      string in `x`.
    """
    f = repr if quoted else str
    y = list(x)
    if len(y) == 1:
        return f(y[0])
    if len(y) == 2:
        return f"{f(y[0])} {c} {f(y[1])}"
    substr = ', '.join(f(i) for i in y[:-1])
    return f"{substr}, {c} {f(y[-1])}"


def allisinstance(types, *targets):
    """True if every target object is an instance of the given type(s).
    
    Parameters
    ----------
    types : type or tuple of types
        The types against which to check the target objects.

    targets
        The objects to check.

    Returns
    -------
    `bool`

    Raises
    ------
    `TypeError`
        The user provided no objects to check.

    See Also
    --------
    `~container.hastype`
        True if an iterable object contains certain types.

    Notes
    -----
    - The calling signature is reversed from that of `isinstance(x, types)` in
      order to accommodate a variable number of target objects.
    - Calling this function with a single target object, `x`, is equivalent to
      calling `isinstance(x, types)`.
    """
    if not targets:
        raise TypeError("Missing object arguments") from None
    return all(isinstance(target, types) for target in targets)


V = typing.TypeVar('V')


class classproperty(typing.Generic[T, V]):
    """A read-only property derived from a `classmethod`.

    Notes
    -----
    - This implementation is based on Django's `classproperty`:
      https://docs.djangoproject.com/en/3.0/_modules/django/utils/decorators/
    - See other suggestions at
        - https://stackoverflow.com/a/13624858/4739101
        - https://stackoverflow.com/a/5191224
    - See https://docs.python.org/3/howto/descriptor.html#properties for a
      pure-Python implementation of `property`.
    - See notes at https://docs.python.org/3/howto/descriptor.html#class-methods
      for additional notes about chaining `classmethod` with other decorators.
    """

    def __init__(
        self,
        method: typing.Callable[[typing.Type[T]], V],
    ) -> None:
        self.fget = classmethod(method)

    def __get__(self, obj: T, objtype: typing.Type[T]=None):
        t = type(obj) if objtype is None else objtype
        return self.fget.__get__(t, t)()

    def getter(self, method: typing.Callable[[typing.Type[T]], V]):
        self.fget = classmethod(method)
        return self


def isproperty(this: object) -> bool:
    """True if `this` is a Python instance property.
    
    This function is designed for use as a predicate with `inspect.getmembers`.
    """
    return inspect.isdatadescriptor(this) and isinstance(this, property)


A = typing.TypeVar('A')
B = typing.TypeVar('B')
T = typing.TypeVar('T')

def widest(a: A, b: B, base: typing.Optional[typing.Type[T]]=None):
    """Determine the widest type between `a` and `b`."""
    if base is None:
        if type(a) == type(b):
            return type(a)
        if issubclass(type(a), type(b)):
            return type(a)
        if issubclass(type(b), type(a)):
            return type(b)
        raise ValueError(
            f"Cannot determine appropriate type from {type(a)} and {type(b)}"
        ) from None
    if not (issubclass(type(a), base) and issubclass(type(b), base)):
        raise TypeError(
            f"{base} is not a common base type of {type(a)} and {type(b)}"
        ) from None
    if isinstance(a, base) and isinstance(b, base):
        if type(a) == type(b):
            return type(a)
        if issubclass(type(a), type(b)):
            return type(a)
        if issubclass(type(b), type(a)):
            return type(b)
        return base
    if isinstance(a, base):
        return type(a)
    if isinstance(b, base):
        return type(b)
    raise ValueError(
        f"Cannot determine appropriate type from {type(a)} and {type(b)}"
    ) from None





"""
Metadata for EPREM simulation parameters.
"""

from .. import aliased


_LOCAL = {
    'minimum_energy': {
        'aliases': ['Emin', 'minimum energy'],
        'unit': 'MeV',
        'default': 0.0,
    },
}
"""Metadata for post-processing parameters."""

# TODO: Add ENLIL-related parameters to _CONFIGURATION_C and _BASETYPES_H

_BASETYPES_H = {
    'AU': {
        'info': 'One astronomical unit.',
        'unit': 'cm',
        'type': float,
    },
    'C': {
        'info': 'The speed of light.',
        'unit': 'cm/s',
        'type': float,
    },
    'DAY': {
        'info': 'The conversion from EPREM time steps to Julian days.',
        'unit': 'day',
        'type': float,
    },
    'DEG2RAD': {
        'info': 'The conversion from degrees to radians.',
        'unit': '1',
        'type': float,
    },
    'EV': {
        'info': 'The conversion from eVs to ergs.',
        'unit': 'erg/eV',
        'type': float,
    },
    'F': {
        'info': 'False',
        'unit': None,
        'type': int,
    },
    'FCONVERT': {
        'info': 'The conversion from distribution to flux.',
        'unit': '1',
        'type': float,
    },
    'GEV': {
        'info': 'The conversion from GeVs to ergs.',
        'unit': 'erg/GeV',
        'type': float,
    },
    'HALF_PI': {
        'info': 'The value of π/2.',
        'unit': None,
        'type': float,
    },
    'LARGEFLOAT': {
        'info': 'A very large float value.',
        'unit': None,
        'type': float,
    },
    'LARGEINT': {
        'info': 'A very large integer value.',
        'unit': None,
        'type': int,
    },
    'MAX_STRING_SIZE': {
        'info': '',
        'unit': None,
        'type': int,
    },
    'MEV': {
        'info': 'The conversion from MeVs to ergs.',
        'unit': 'erg/MeV',
        'type': float,
    },
    'MHD_B_NORM': {
        'info': 'The normalization for magnetic fields.',
        'unit': '1',
        'type': float,
    },
    'MHD_COUPLED': {
        'info': 'Couple to external MHD',
        'unit': None,
        'type': int,
    },
    'MHD_DEFAULT': {
        'info': 'Use the default MHD solver.',
        'unit': None,
        'type': int,
    },
    'MHD_DENSITY_NORM': {
        'info': 'The normalization factor for density.',
        'unit': '1',
        'type': float,
    },
    'MP': {
        'info': 'The proton mass.',
        'unit': 'g',
        'type': float,
    },
    'MZERO': {
        'info': 'The proton rest-mass energy in GeV.',
        'unit': 'GeV',
        'type': float,
    },
    'OM': {
        'info': 'The normalization for ion gyrofrequency.',
        'unit': '1',
        'type': float,
    },
    'PI': {
        'info': 'The value of π.',
        'unit': None,
        'type': float,
    },
    'Q': {
        'info': 'The proton charge.',
        'unit': 'statC',
        'type': float,
    },
    'RAD2DEG': {
        'info': 'The conversion from radians to degrees.',
        'unit': '1',
        'type': float,
    },
    'RSAU': {
        'info': 'The number of solar radii per au.',
        'unit': '1',
        'type': float,
    },
    'RSKM': {
        'info': 'The number of solar radii per km.',
        'unit': '1',
        'type': float,
    },
    'RSUN': {
        'info': 'The value of the solar radius.',
        'unit': 'cm',
        'type': float,
    },
    'SMALLFLOAT': {
        'info': 'A very small float value.',
        'unit': None,
        'type': float,
    },
    'T': {
        'info': 'True',
        'unit': None,
        'type': int,
    },
    'TAU': {
        'info': 'The canonical EPREM time scale.',
        'unit': 's',
        'type': float,
    },
    'THRESH': {
        'info': 'The threshold for perpendicular diffusion.',
        'unit': '1',
        'type': float,
    },
    'TWO_PI': {
        'info': 'The value of 2π.',
        'unit': None,
        'type': float,
    },
    'VOLT': {
        'info': 'The conversion from volts to statvolts.',
        'unit': '1',
        'type': float,
    },
}
"""Metadata for parameters defined in `baseTypes.h`."""

_CONFIGURATION_C = {
    'FailModeDump': {
        'aliases': [],
        'unit': None,
    },
    'abundance': {
        'aliases': [],
        'unit': None,
    },
    'adiabaticChangeAlg': {
        'aliases': [],
        'unit': None,
    },
    'adiabaticFocusAlg': {
        'aliases': [],
        'unit': None,
    },
    'aziSunStart': {
        'aliases': [],
        'unit': 'rad',
    },
    'boundaryFunctAmplitude': {
        'aliases': ['J0'],
        'unit': '1 / (cm^2 * s * sr * (MeV/nuc))',
    },
    'boundaryFunctBeta': {
        'aliases': ['beta'],
        'unit': '1',
    },
    'boundaryFunctEcutoff': {
        'aliases': ['E0'],
        'unit': 'MeV',
    },
    'boundaryFunctEr': {
        'aliases': ['Er', 'reference energy'],
        'unit': 'MeV',
    },
    'boundaryFunctGamma': {
        'aliases': ['gamma'],
        'unit': '1',
    },
    'boundaryFunctR0': {
        'aliases': ['R0', 'r0', 'reference radius'],
        'unit': 'au',
    },
    'boundaryFunctXi': {
        'aliases': ['xi'],
        'unit': '1',
    },
    'boundaryFunctionInitDomain': {
        'aliases': [],
        'unit': None,
    },
    'charge': {
        'aliases': [],
        'unit': 'e',
    },
    'checkSeedPopulation': {
        'aliases': [],
        'unit': None,
    },
    'dumpFreq': {
        'aliases': [],
        'unit': None,
    },
    'eMax': {
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'eMin': {
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'epCalcStartTime': {
        'aliases': [],
        'unit': 'day',
    },
    'epEquilibriumCalcDuration': {
        'aliases': [],
        'unit': 'day',
    },
    'epremDomain': {
        'aliases': [],
        'unit': None,
    },
    'epremDomainOutputTime': {
        'aliases': [],
        'unit': 'day',
    },
    'flowMag': {
        'aliases': [],
        'unit': 'cm/s',
    },
    'fluxLimiter': {
        'aliases': [],
        'unit': None,
    },
    'gammaEhigh': {
        'aliases': [],
        'unit': '1',
    },
    'gammaElow': {
        'aliases': [],
        'unit': '1',
    },
    'idealShock': {
        'aliases': [],
        'unit': None,
    },
    'idealShockFalloff': {
        'aliases': [],
        'unit': None,
    },
    'idealShockInitTime': {
        'aliases': [],
        'unit': 'day',
    },
    'idealShockJump': {
        'aliases': [],
        'unit': '1',
    },
    'idealShockPhi': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockPhiWidth': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockScale': {
        'aliases': [],
        'unit': 'au',
    },
    'idealShockScaleLength': {
        'aliases': [],
        'unit': 'au',
    },
    'idealShockSharpness': {
        'aliases': [],
        'unit': '1',
    },
    'idealShockSpeed': {
        'aliases': [],
        'unit': 'cm/s',
    },
    'idealShockTheta': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockThetaWidth': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockWidth': {
        'aliases': [],
        'unit': 'rad',
    },
    'idw_p': {
        'aliases': [],
        'unit': '1',
    },
    'kperxkpar': {
        'aliases': ['kper_kpar', 'kper/kpar', 'kper / kpar'],
        'unit': '1',
    },
    'lamo': {
        'aliases': ['lam0', 'lambda0'],
        'unit': 'au',
    },
    'mass': {
        'aliases': [],
        'unit': 'nucleon',
    },
    'mfpRadialPower': {
        'aliases': ['mfp_radial_power'],
        'unit': '1',
    },
    'mhdBAu': {
        'aliases': [],
        'unit': 'G',
    },
    'mhdBConvert': {
        'aliases': [],
        'unit': '1',
    },
    'mhdCouple': {
        'aliases': [],
        'unit': None,
    },
    'mhdCoupledTime': {
        'aliases': [],
        'unit': None,
    },
    'mhdDigits': {
        'aliases': [],
        'unit': None,
    },
    'mhdDensityAu': {
        'aliases': [],
        'unit': 'cm^-3',
    },
    'mhdDirectory': {
        'aliases': [],
        'unit': None,
    },
    'mhdInitFromOuterBoundary': {
        'aliases': [],
        'unit': None,
    },
    'mhdInitRadius': {
        'aliases': [],
        'unit': 'au',
    },
    'mhdInitTimeStep': {
        'aliases': [],
        'unit': 'day',
    },
    'mhdNumFiles': {
        'aliases': [],
        'unit': None,
    },
    'mhdRadialMax': {
        'aliases': [],
        'unit': 'au',
    },
    'mhdRadialMin': {
        'aliases': [],
        'unit': 'au',
    },
    'mhdRhoConvert': {
        'aliases': [],
        'unit': '1',
    },
    'mhdRotateSolution': {
        'aliases': [],
        'unit': None,
    },
    'mhdStartTime': {
        'aliases': [],
        'unit': 'day',
    },
    'mhdSteadyState': {
        'aliases': [],
        'unit': None,
    },
    'mhdTimeConvert': {
        'aliases': [],
        'unit': '1',
    },
    'mhdVConvert': {
        'aliases': [],
        'unit': '1',
    },
    'mhdVmin': {
        'aliases': [],
        'unit': 'cm/s',
    },
    'mhdVmax': {
        'aliases': [],
        'unit': 'cm/s',
    },
    'minInjectionEnergy': {
        'aliases': [],
        'unit': 'MeV',
    },
    'numColumnsPerFace': {
        'aliases': [],
        'unit': None,
    },
    'numEnergySteps': {
        'aliases': [],
        'unit': None,
    },
    'numEpSteps': {
        'aliases': [],
        'unit': None,
    },
    'numMuSteps': {
        'aliases': [],
        'unit': None,
    },
    'numNodesPerStream': {
        'aliases': [],
        'unit': None,
    },
    'numObservers': {
        'aliases': [],
        'unit': '1',
    },
    'numRowsPerFace': {
        'aliases': [],
        'unit': None,
    },
    'numSpecies': {
        'aliases': [],
        'unit': None,
    },
    'obsPhi': {
        'aliases': [],
        'unit': 'rad',
    },
    'obsR': {
        'aliases': [],
        'unit': 'au',
    },
    'obsTheta': {
        'aliases': [],
        'unit': 'rad',
    },
    'omegaSun': {
        'aliases': [],
        'unit': 'rad * cm / (au * s)',
    },
    'outputFloat': {
        'aliases': [],
        'unit': None,
    },
    'outputFlux': {
        'aliases': [],
        'unit': None,
    },
    'pointObserverOutput': {
        'aliases': [],
        'unit': None,
    },
    'pointObserverOutputTime': {
        'aliases': [],
        'unit': 'day',
    },
    'preEruptionDuration': {
        'aliases': [],
        'unit': 'day',
    },
    'rScale': {
        'aliases': [],
        'unit': 'au',
    },
    'rigidityPower': {
        'aliases': ['rigidity_power'],
        'unit': '1',
    },
    'shockDetectPercent': {
        'aliases': [],
        'unit': '1',
    },
    'shockInjectionFactor': {
        'aliases': [],
        'unit': '1',
    },
    'shockSolver': {
        'aliases': [],
        'unit': None,
    },
    'simStartTime': {
        'aliases': [],
        'unit': 'day',
    },
    'simStopTime': {
        'aliases': [],
        'unit': 'day',
    },
    'streamFluxOutput': {
        'aliases': [],
        'unit': None,
    },
    'streamFluxOutputTime': {
        'aliases': [],
        'unit': 'day',
    },
    'streamSpawnLocAzi': {
        'aliases': [],
        'unit': 'rad',
    },
    'streamSpawnLocZen': {
        'aliases': [],
        'unit': 'rad',
    },
    'subTimeCouple': {
        'aliases': [],
        'unit': None,
    },
    'tDel': {
        'aliases': [],
        'unit': 'day',
    },
    'unifiedOutput': {
        'aliases': [],
        'unit': None,
    },
    'unifiedOutputTime': {
        'aliases': [],
        'unit': 'day',
    },
    'unstructuredDomain': {
        'aliases': [],
        'unit': None,
    },
    'unstructuredDomainOutputTime': {
        'aliases': [],
        'unit': 'day',
    },
    'useAdiabaticChange': {
        'aliases': [],
        'unit': None,
    },
    'useAdiabaticFocus': {
        'aliases': [],
        'unit': None,
    },
    'useDegrees': {
        'aliases': [],
        'unit': None,
    },
    'useDrift': {
        'aliases': [],
        'unit': None,
    },
    'useManualStreamSpawnLoc': {
        'aliases': [],
        'unit': None,
    },
    'useMhdSteadyStateDt': {
        'aliases': [],
        'unit': None,
    },
    'useParallelDiffusion': {
        'aliases': [],
        'unit': None,
    },
    'useShellDiffusion': {
        'aliases': [],
        'unit': None,
    },
    'warningsFile': {
        'aliases': [],
        'unit': None,
    },
}
"""Metadata for parameters defined in `configuration.c`."""

_metadata = {**_LOCAL, **_BASETYPES_H, **_CONFIGURATION_C}
"""Combined metadata dictionary for internal use."""

METADATA = aliased.Mapping(_metadata)
"""Metadata for EPREM parameters."""

NAMES = aliased.NameMap(_metadata)
"""A mapping from alias(es) to EPREM parameter."""

ALIASES = aliased.Groups(*METADATA.keys(aliased=True))
"""Groups of aliases for EPREM parameters."""

UNITS = {
    key: info.get('unit')
    for key, info in _metadata.items()
}
"""Collection of units from metadata."""





import abc
import typing

from ._units import Unit


T = typing.TypeVar('T')


class Interface(typing.Protocol[T]):
    """Abstract protocol class for metric objects.

    Objects that define the following properties are compatible with this
    protocol

    - `isunitless`, which should return a `bool`
    - `unit`, which should return a `~metric.Unit`
    """

    @property
    @abc.abstractmethod
    def isunitless(self: T) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def unit(self: T) -> Unit:
        raise NotImplementedError


@typing.runtime_checkable
class Type(typing.Protocol[T]):
    """Structural protocol class for metric objects.

    Objects that define the following properties will pass instance checks
    against this protocol

    - `isunitless`, which should return a `bool`
    - `unit`, which should return a `~metric.Unit`
    """

    @property
    def isunitless(self: T) -> bool: ...

    @property
    def unit(self: T) -> Unit: ...






import collections.abc
import contextlib
import numbers
import typing

import numpy
import numpy.typing

from .. import base
from .. import container
from .. import etc
from . import _data
from . import _index
from . import _operators
from ..typehelp import Self
from ._objects import (
    Object,
    Quantity,
)
from ._mixins import Operators


T = typing.TypeVar('T')


class DimensionTypeError(Exception):
    """Invalid type for operation on dimensions."""


@etc.autostr
class Dimensions(collections.abc.Sequence):
    """A representation of one or more array-axis names.
    
    This class is formally a sequence but it supports certain set-like
    operations in an order-preserving way.
    """

    def __init__(self, names: typing.Tuple[str]) -> None:
        self._names = names

    def __eq__(self, other):
        """Called for self == other."""
        if isinstance(other, Dimensions):
            return self._names == other._names
        if isinstance(other, str) and len(self) == 1:
            return self._names[0] == other
        try:
            truth = self._names == tuple(other)
        except TypeError:
            return False
        return truth

    def __lt__(self, other):
        """Called for self < other."""
        return self._compare(other, _operators.lt)

    def __le__(self, other):
        """Called for self <= other."""
        return self._compare(other, _operators.le)

    def __gt__(self, other):
        """Called for self > other."""
        return self._compare(other, _operators.gt)

    def __ge__(self, other):
        """Called for self >= other."""
        return self._compare(other, _operators.ge)

    def _compare(self, other, operator: typing.Callable[..., T]) -> T:
        """Implement a binary comparison between this instance and `other`."""
        these = set(self._names)
        those = set(other._names if isinstance(other, Dimensions) else other)
        return operator(these, those)

    def replace(self, old: str, new: str):
        """Replace a single dimension."""
        return dimensions_factory(
            *[new if n == old else n for n in self._names]
        )

    def insert(self, name: str, index: typing.SupportsIndex=None):
        """Insert the named dimension at `index`."""
        names = list(self._names)
        if index is not None:
            names.insert(index, name)
        else:
            names.append(name)
        return dimensions_factory(*names)

    def __and__(self, other):
        """Called for self & other."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        full = list(self._names) + those
        common = set(self._names) & set(those)
        return dimensions_factory(*[n for n in full if n in common])

    def __rand__(self, other):
        """Called for other & self."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        full = those + list(self._names)
        common = set(self._names) & set(those)
        return dimensions_factory(*[n for n in full if n in common])

    def __sub__(self, other):
        """Called for self - other."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        return dimensions_factory(*[n for n in self if n not in those])

    def __rsub__(self, other):
        """Called for other - self."""
        if isinstance(other, Dimensions):
            return dimensions_factory(*[n for n in other if n not in self])
        raise NotImplemented

    def __or__(self, other):
        """Called for self | other."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        these = list(self._names)
        try:
            merged = container.merge(these, those)
        except container.MergeError as err:
            errmsg = str(err).replace('entries', 'dimensions')
            raise ValueError(errmsg) from err
        return dimensions_factory(merged)

    def __ror__(self, other):
        """Called for other | self."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        these = list(self._names)
        try:
            merged = container.merge(those, these)
        except container.MergeError as err:
            errmsg = str(err).replace('entries', 'dimensions')
            raise ValueError(errmsg) from err
        return dimensions_factory(merged)

    def __xor__(self, other):
        """Called for self ^ other."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        return (self - those) | (those - self)

    def __rxor__(self, other):
        """Called for other ^ self."""
        try:
            those = self._get_operand(other)
        except DimensionTypeError:
            return NotImplemented
        return (those - self) | (self - those)

    def _get_operand(self, arg) -> typing.List[str]:
        """Return an appropriate operand for binary operations."""
        if isinstance(arg, Dimensions):
            return list(arg._names)
        if isinstance(arg, str):
            return [arg]
        try:
            iter(arg)
        except TypeError:
            raise DimensionTypeError(arg)
        if all(isinstance(i, str) for i in arg):
            return list(arg)
        raise DimensionTypeError(arg)

    def copy(self):
        """Create a copy of this instance."""
        return dimensions_factory(*self._names)

    @typing.overload
    def permute(
        self,
        order: typing.Iterable[typing.SupportsIndex],
    ) -> Self: ...

    @typing.overload
    def permute(self, *order: typing.SupportsIndex) -> Self: ...

    def permute(self, *args, **kwargs):
        """Reorder these dimensions."""
        if args and 'order' in kwargs:
            raise TypeError("Too many arguments") from None
        if args:
            if len(args) == 1:
                return self._permute(args[0])
            return self._permute(args)
        if 'order' in kwargs:
            return self._permute(kwargs['order'])
        raise TypeError("No arguments") from None

    def _permute(self, axes: typing.Iterable[typing.SupportsIndex]):
        """Helper for `~permute`."""
        try:
            names = [self[axis] for axis in axes]
        except (IndexError, TypeError) as err:
            raise ValueError(
                "Cannot interpret all values as indices"
            ) from err
        self.mismatch(names)
        return dimensions_factory(names)

    def mismatch(self, other: typing.Iterable[str]):
        """Raise an exception if `self < other` or `self > other`.

        This is a convenience method to checking whether `self == other` in a
        set-wise sense, and issuing a descriptive message depending upon whether
        `self` is a subset of `other` or vice versa. Note that it can also be
        used as a static method to compare two iterables of strings.

        Parameters
        ----------
        other : iterable of strings
            The object to compare to this collection (or to the first argument,
            when called as a static method).

        Returns
        -------
        `None`
            The two collections are equal in a set-wise sense.

        Raises
        ------
        `ValueError`
            The two collections are unequal in a set-wise sense.
        """
        a = set(self)
        b = set(other)
        if a < b:
            raise ValueError("Not enough dimensions") from None
        if a > b:
            diff = a - b
            dims = etc.join(diff, 'or', quoted=True)
            raise ValueError(
                f"Original dimensions do not include {dims}"
            ) from None

    def __hash__(self):
        """Support use as a mapping key."""
        return hash(self._names)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._names)

    def __getitem__(self, __i: typing.SupportsIndex):
        """Called for index-based access."""
        return self._names[__i]

    def __str__(self) -> str:
        return f"{{{', '.join(repr(name) for name in self._names)}}}"


@typing.overload
def dimensions_factory(*args) -> Dimensions:
    """Create array dimensions from the given arguments, if possible."""

@typing.overload
def dimensions_factory(names: typing.Iterable[str]) -> Dimensions:
    """Create array dimensions from an iterable of names."""

@typing.overload
def dimensions_factory(*names: str) -> Dimensions:
    """Create array dimensions from one or more names."""

@typing.overload
def dimensions_factory(n: int) -> Dimensions:
    """Create `n` array dimensions with default names."""

def dimensions_factory(*args):
    """Factory function for data dimensions."""
    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, numbers.Real):
            return tuple(f'x{i}' for i in range(arg))
    unwrapped = container.unwrap(args, newtype=tuple)
    unique = container.unique(unwrapped)
    return Dimensions(_compute_names(unique))


def _compute_names(args: typing.Iterable[str]) -> typing.Tuple[str]:
    """Determine the instance names from arguments."""
    if isinstance(args, Dimensions):
        return args._names
    if all(isinstance(arg, str) for arg in args):
        return tuple(args)
    raise TypeError(
        f"Unable to compute dimension names from {args!r}"
    ) from None


IndexLike = typing.Union[
    typing.SupportsIndex,
    typing.Tuple[typing.SupportsIndex],
    slice,
]

ArrayLike = typing.Union[base.Array, numpy.typing.NDArray]

DataType = typing.TypeVar('DataType', base.Array, numpy.ndarray)


class Array(Quantity[DataType], Operators):
    """Base class for numeric arrays with named dimensions."""

    def __init__(
        self,
        data: DataType,
        dimensions: Dimensions,
    ) -> None:
        super().__init__(data)
        self._shape = data.shape
        self._size = data.size
        self._ndim = data.ndim
        self._dimensions = dimensions
        self._array = data if isinstance(data, numpy.ndarray) else None

    def __eq__(self, other):
        """Called for self == other."""
        if isinstance(other, Array):
            return (
                numpy.array_equal(self.array, other.array)
                and
                self.dimensions == other.dimensions
            )
        return False

    def __ne__(self, other):
        """Called for self != other."""
        return not (self == other)

    def __contains__(self, __x) -> bool:
        """Called for x in self.
        
        See Also
        --------
        array_contains
            Perform additional containment checks beyond this method.
        """
        return __x in self._object or __x in self.array

    def __iter__(self):
        """Called for iter(self)."""
        return iter(self._object)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._object)

    def __getitem__(self, args: IndexLike):
        """Called for subscription."""
        x = [args] if isinstance(args, typing.SupportsIndex) else args
        return self.spawn(
            self._getitem(x),
            dimensions=self.dimensions,
        )

    def _getitem(self, args):
        """Extract a subarray.

        This method exists independently of `__getitem__` in order to separate
        the subscription logic from the return-type logic.
        """
        unwrapped = container.unwrap(args)
        indices = _index.resolve(self.shape, unwrapped)
        loaded = self._load_array(indices)
        array = numpy.array(loaded, ndmin=self.ndim)
        if isinstance(indices, numpy.ndarray) and indices.dtype == bool:
            return array
        if loaded.ndim != self.ndim:
            with contextlib.suppress(TypeError):
                shape = [
                    self._get_axis_size(i, v)
                    for i, v in enumerate(indices)
                ]
                return array.reshape(shape)
        return array

    def _get_axis_size(self, i: int, v):
        """Helper for computing shape in `__getitem__`."""
        if isinstance(v, int):
            return 1
        if isinstance(v, slice):
            return (v.stop or self.shape[i]) - (v.start or 0)
        return self.shape[i]

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self._as_string()

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return self._as_string(
            prefix=f'{self.__class__.__qualname__}(',
            suffix=')',
        )

    def _as_string(self, prefix: str='', suffix: str='') -> str:
        """Create a string representation of this object."""
        data = self._format_data(prefix=prefix, suffix=suffix)
        dimensions = f"dimensions={self.dimensions}"
        return f"{prefix}{data}, {dimensions}{suffix}"

    def _format_data(self, **kwargs):
        """Create an appropriate string to represent the data."""
        if not isinstance(self._object, numpy.ndarray):
            return str(type(self._object))
        try:
            signed = '+' if any(i < 0 for i in self.array.flat) else '-'
        except TypeError:
            # If we get here, it's probably because some values are strings
            signed = '-'
        return numpy.array2string(
            self.array,
            threshold=4,
            edgeitems=2,
            separator=', ',
            sign=signed,
            precision=3,
            floatmode='maxprec_equal',
            **kwargs
        )

    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Support casting to `numpy` array types.
        
        Notes
        -----
        This will retrieve the underlying array before applying `*args` and
        `**kwargs`, in order to avoid a `TypeError` when using
        `netCDF4.Dataset`. See
        https://github.com/mcgibbon/python-examples/blob/master/scripts/file-io/load_netCDF4_full.py
        """
        array = self.array
        return numpy.asarray(array, *args, **kwargs)

    def remesh(self, other, invert: bool=False):
        """Construct array meshes for broadcasting.

        This method calls `~remesh` to reshape the arrays of both data
        interfaces in a way that is consistent with their combined dimensions.

        Parameters
        ----------
        other : `~Data`
            The data interface with which to mutually reshape.
        invert : bool, default=false
            If true, reverse the order of `self` and `other`.
        """
        if invert:
            return remesh(other, self)
        return remesh(self, other)

    def contains(self, value: numbers.Real) -> bool:
        """True if `value` is in this sequence's data array.

        This method uses `~numeric.data.isclose` to test whether `value` is
        equal, or very close, to a value in this object's real-valued data.

        Parameters
        ----------
        value : real
            The value for which to search.
        
        See Also
        --------
        `~numeric.data.isclose`
        """
        return _data.isclose(self, value)

    @typing.overload
    def transpose(
        self,
        axes: typing.Iterable[typing.Union[str, typing.SupportsIndex]],
        /,
    ) -> Self: ...

    @typing.overload
    def transpose(
        self,
        *axes: typing.Union[str, typing.SupportsIndex],
    ) -> Self: ...

    def transpose(self, *args):
        """Transpose this array."""
        if not args:
            return self
        if etc.allisinstance(str, *args):
            return self._transpose(args)
        if len(args) == 1:
            arg = args[0]
            if etc.allisinstance(str, *arg):
                return self._transpose(arg)
        data = self.array.transpose(*args)
        new = self.dimensions.permute(*args)
        return self.spawn(data, dimensions=new)

    def _transpose(self, args: typing.Iterable[str]):
        """Helper for `~transpose`."""
        self.dimensions.mismatch(args)
        axes = [self.dimensions.index(arg) for arg in args]
        data = self.array.transpose(axes)
        return self.spawn(data, dimensions=args)

    @property
    def size(self) -> int:
        """The number of elements in this array."""
        return self._size

    @property
    def array(self):
        """The underlying array-like data interface."""
        if self._array is None:
            self._array = self._load_array()
        return self._array

    def _load_array(
        self,
        index: typing.Optional[IndexLike]=None,
    ) -> numpy.typing.NDArray:
        """Load the array from disk or memory.

        This method is intended for internal use. Users should subscript
        instances via bracket notation (e.g., `array[:, 2]`).

        If `index` is `None` or an empty iterable, this method will produce the
        entire array. Otherwise, it will create the requested subarray from the
        internal `_data` attribute. It will always attempt to use a cached
        version of the full array before loading from disk. Therefore,
        repeatedly calling this method should not create a performance
        bottleneck.

        The specific algorithm is as follows:

        - If `index` is null (i.e., `None` or an empty iterable object, but not
          literal 0), the caller wants the full array.
        
            - If we already have the full array, return it.
            - Else, read it, save it, and return it.

        - Else, if `index` is not null, the caller wants a subarray.

            - If we already have the full array, subscript and return it.
            - Else, continue

        - Else, read and subscript the array, and return the subarray.

        The reasoning behind this algorithm is as follows: If we need to load
        the full array at any point, we may as well save it because subscripting
        an in-memory `numpy.ndarray` is much faster than re-reading from disk
        for large arrays. However, we should avoid reading in the full array if
        the caller only wants a small portion of it. We don't cache these
        subarrays because reusing a subarray is only meaningful if the indices
        haven't changed. Furthermore, accessing a subarray via bracket syntax
        creates a new object, at which point the subarray becomes the new
        object's full array.
        """
        if etc.isnull(index):
            if self._array is not None:
                return self._array
            array = self._read_array()
            self._array = array
            return array
        if self._array is not None:
            idx = numpy.index_exp[index]
            return self._array[idx]
        return self._read_array(index)

    def _read_array(
        self,
        index: typing.Optional[IndexLike]=None,
    ) -> numpy.typing.NDArray:
        """Read the array data from disk.
        
        If `index` is null in the sense defined by `~etc.isnull` this method
        will load and return the full array. If `index` is not null, this method
        will first attempt to subscript the internal `_data` attribute before
        converting it to an array and returning it. If it catches either a
        `TypeError` or an `IndexError`, it will create the full array before
        subscripting and returning it. The former may occur if the internal
        `_data` attribute is a sequence type like `list`, `tuple`, or `range`;
        the latter may occur when attempting to subscript certain array-like
        objects (e.g., `netCDF4._netCDF4.Variable`) despite passing a valid
        `numpy` index expression.
        """
        if not etc.isnull(index):
            idx = numpy.index_exp[index]
            try:
                return numpy.asarray(self._object[idx])
            except (TypeError, IndexError):
                return numpy.asarray(self._object)[idx]
        return numpy.asarray(self._object)

    @property
    def shapemap(self) -> typing.Dict[str, int]:
        """The name and size of each axis."""
        return dict(zip(self.dimensions, self.shape))

    @property
    def dimensions(self) -> Dimensions:
        """The names of this array's indexable axes."""
        return self._dimensions

    @property
    def ndim(self) -> int:
        """The number of dimensions in this array."""
        return self._ndim

    @property
    def shape(self) -> typing.Tuple[int]:
        """The length of each dimension in this array."""
        return self._shape

    def _get_arg_data(self, arg):
        if isinstance(arg, Array):
            return self.array
        return super()._get_arg_data(arg)

    _FUNCTIONS = {}
    _OPERATORS = {}


Array._UFUNC_TYPES += [Array]
Array._FUNCTION_TYPES += [Array]


ArrayType = typing.TypeVar('ArrayType', bound=Array)


@Array.implements(numpy.array_equal)
def array_equal(a: Array, b: Array) -> bool:
    """Called for numpy.array_equal(a, b)."""
    return numpy.array_equal(numpy.array(a), numpy.array(b))


@typing.overload
def array_factory(*args, **kwargs) -> Array:
    """Create an array-like quantity from the given arguments, if possible."""

@typing.overload
def array_factory(x, /) -> Array:
    """Create an array-like quantity from `x`.

    This function will construct default dimension names based on the number of
    dimensions in `x`.
    """

@typing.overload
def array_factory(x, /, dimensions: typing.Iterable[str]) -> Array:
    """Create an array-like quantity from `x`.

    This function will use the strings in `dimensions` as the dimension names.
    The length of `dimensions` must equal the number of dimensions in `x`.
    """

@typing.overload
def array_factory(x, *dimensions: str) -> Array:
    """Create an array-like quantity from `x`.

    This function will use the string arguments following `x` as the dimension
    names. The number of strings must equal the number of dimensions in `x`.
    """

@typing.overload
def array_factory(x: ArrayType, /) -> ArrayType:
    """Create an array-like quantity from `x`.

    This function will copy array data and dimensions from `x`.
    """

def array_factory(*args, **kwargs):
    return Array(*array_args(*args, **kwargs))

def array_args(*args, **kwargs):
    if len(args) == 1:
        x = args[0]
        if isinstance(x, Array):
            if kwargs:
                raise TypeError(
                    "Cannot change attributes when copying"
                    " an existing quantity"
                ) from None
            return x.array, x.dimensions
    array, dimensions = _parse_data_args(*args, **kwargs)
    if len(dimensions) != array.ndim:
        raise ValueError(
            f"Number of named dimensions ({len(dimensions)})"
            f" must equal number of array dimensions ({array.ndim})"
        ) from None
    return array, dimensions


def _parse_data_args(*args, dimensions=None):
    """Parse arguments for initializing a data interface."""
    array = _init_array(args[0])
    if len(args) > 1:
        if dimensions:
            raise TypeError(
                "Got positional and keyword arguments for dimensions"
            ) from None
        return array, dimensions_factory(args[1:])
    if dimensions is None:
        return array, dimensions_factory(array.ndim)
    return array, dimensions_factory(dimensions)


def _init_array(arg) -> base.Array:
    """Get an array-like data interface.
    
    This method ensures that calling methods get an object that has certain
    array-like properties, without requiring that `arg` be an instance of
    `numpy.ndarray` and without converting `arg` to a `numpy.ndarray` unless
    necessary.
    """
    data = _get_arraylike(arg)
    # NOTE: This attribute check is more robust than testing whether `data`
    # satisfies the `quantity.Array` protocol, which requires that the target
    # object define these attribute as properties.
    if all(hasattr(data, name) for name in ('shape', 'ndim', 'size')):
        return data
    return numpy.array(data)


def _get_arraylike(arg):
    """Extract an array-like interface from `arg`."""
    if isinstance(arg, Object):
        return arg.data
    if isinstance(arg, (list, tuple, set)):
        return numpy.array(arg)
    return arg


def remesh(a: Array, b: Array):
    """Construct array meshes for broadcasting.
    
    This function mutually reshapes the internal `numpy` arrays of two data
    interfaces based on their combined dimensions. The result is a pair of
    `numpy` arrays that can broadcast together while preserving a shape
    consistent with the combined dimensions.

    Parameters
    ----------
    a, b : `~numeric.Array`
        The numeric arrays to reshape.

    Returns
    -------
    `tuple` of `numpy.ndarray`
        The reshaped `numpy` array of `a`, followed by that of `b`.

    Raises
    ------
    `TypeError`
        One or both of the arguments was not an instance of `~numeric.Array`.
    `ValueError`
        The original array shapes of `a` and `b` have incompatible shapes and
        dimensions. Note that this case acts like the `ValueError` raised when
        attempting to broadcast two `numpy` arrays with inconsistent shapes.
    """
    if not (isinstance(a, Array) and isinstance(b, Array)):
        raise TypeError(
            f"Both arguments must be instances of {Array!r}"
        ) from None
    dimensions = a.dimensions | b.dimensions
    for d in dimensions:
        adim = a.shapemap.get(d, 1)
        bdim = b.shapemap.get(d, 1)
        if adim > 1 and bdim > 1 and adim != bdim:
            # NOTE: This acts like the `ValueError` raised when attempting
            # to broadcast two `numpy` arrays with inconsistent shapes.
            raise ValueError(
                "Cannot create a mesh from arrays with"
                f" dimensions {a.dimensions} and {b.dimensions}"
                ", and corresponding shapes"
                f" shapes {a.shape} and {b.shape}"
            )
    ndim = len(dimensions)
    shape = tuple(
        # NOTE: This creates the maximal shape in order to handle cases in
        # which the input arrays have matching dimensions but one array is
        # singular. The previous implementation would always overwrite the
        # size of matching dimensions in `a` with those of `b`.
        max(a.shapemap.get(d, 1), b.shapemap.get(d, 1))
        for d in dimensions
    )
    idxmesh = numpy.ix_(*[range(i) for i in shape])
    idxa = tuple(idxmesh[dimensions.index(d)] for d in a.shapemap)
    idxb = tuple(idxmesh[dimensions.index(d)] for d in b.shapemap)
    x = numpy.array(a, ndmin=ndim) if a.size == 1 else a._load_array(idxa)
    y = numpy.array(b, ndmin=ndim) if b.size == 1 else b._load_array(idxb)
    return x, y





"""
The interactive EPREM runtime interface.
"""

import argparse
import collections.abc
import contextlib
import datetime
import json
import os
import pathlib
import shutil
import subprocess
import sys
import typing

try:
    import isort
    _HAVE_ISORT = True
except ModuleNotFoundError:
    _HAVE_ISORT = False
try:
    import yaml
    _HAVE_YAML = True
except ModuleNotFoundError:
    _HAVE_YAML = False

FILEPATH = pathlib.Path(__file__)
DIRECTORY = FILEPATH.parent
sys.path.append(str(DIRECTORY))

from . import interfaces
from . import etc
from . import paths


class LogKeyError(KeyError):
    """The log has no entry for a given key."""


class ReadTypeError(Exception):
    """There is no support for reading a given file type."""


RunLogType = typing.TypeVar('RunLogType', bound='RunLog')


class RunLog(collections.abc.Mapping):
    """Mapping-based interface to an EPREM project log."""

    def __init__(self, path: paths.PathLike, **common) -> None:
        """Create a new project log.
        
        Parameters
        ----------
        path : path-like
            The path to the log file, including the file name. If the file does
            not exist, this class will create a new one; otherwise, this class
            will use the existing file. The path may be relative to the current
            directory.

        common
            Key-value pairs of attributes that are common to all runs. Not
            allowed when retrieving an existing file.

        Raises
        ------
        TypeError
            Caller attempted to set common attributes on an existing log.
        """
        full = paths.normalize(path)
        if full.exists():
            self._path = full
            if common:
                raise TypeError(
                    "Cannot change common attributes "
                    "of an existing log"
                ) from None
        else:
            self._path = full
            self.dump(common)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._asdict)

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(self._asdict)

    def __getitem__(self, __k: str):
        """Get metadata for a run."""
        if __k in self._asdict:
            return self._asdict[__k]
        raise LogKeyError(f"Unknown run {__k!r}")

    def update_directory(self, target: paths.PathLike):
        """Update the directory path to this file.

        Parameters
        ----------
        target : path-like
            The directory to which to move this file. May be relative.

        Raises
        ------
        PathTypeError
            The target path does not exist or is not a directory.

        Notes
        -----
        * This method assumes that the target path exists.
        * This method does not allow changes to the file name.
        """
        new = paths.normalize(target)
        if not new.exists():
            raise paths.PathTypeError("The new path must exist") from None
        if not new.is_dir():
            raise paths.PathTypeError("Cannot rename log file") from None
        self._path = new / self.name
        source = str(self.path.parent)
        for key in self:
            self.mv(key, key.replace(source, str(new)))
        return self

    @typing.overload
    def create(
        self: RunLogType,
        key: str,
        mapping: typing.Mapping
    ) -> RunLogType:
        """Create a new entry from a mapping."""

    @typing.overload
    def create(
        self: RunLogType,
        key: str,
        **items
    ) -> RunLogType:
        """Create a new entry from key-value pairs."""

    def create(
        self: RunLogType,
        key: paths.PathLike,
        *mapping: typing.Mapping,
        **items
    ) -> RunLogType:
        contents = self._asdict.copy()
        if mapping and items:
            raise TypeError(
                f"{self.__class__.__qualname__}.create accepts"
                " a single mapping or multiple key-value pairs"
                ", but not both"
            ) from None
        new = dict(*mapping) or items
        contents[str(key)] = new
        self.dump(contents)
        return self

    def load(
        self: RunLogType,
        key: paths.PathLike,
        source: paths.PathLike,
        filetype: str=None
    ) -> RunLogType:
        """Create a new entry in this log file from a file."""
        contents = self._asdict.copy()
        contents[str(key)] = self._read_from_file(source, filetype)
        self.dump(contents)
        return self

    def _read_from_file(self, source, filetype):
        """Update the current contents from the `source` file."""
        contents = self._asdict.copy()
        loader = self._source_loader(filetype or pathlib.Path(source).suffix)
        with pathlib.Path(source).open('r') as fp:
            if loaded := loader(fp):
                contents.update(loaded)
        return contents

    def _source_loader(self, filetype: str):
        """Get a format-specific file-reader."""
        if filetype.lower().lstrip('.') == 'yaml':
            if _HAVE_YAML:
                return yaml.safe_load
            raise ReadTypeError("No support for reading YAML") from None
        if filetype.lower().lstrip('.') == 'json':
            return json.load
        raise ValueError(
            f"Unknown file type: {filetype!r}"
        ) from None

    def append(self, target: paths.PathLike, key: str, value):
        """Append metadata to `target`."""
        contents = self._asdict.copy()
        run = str(target)
        try:
            record = contents[run]
        except KeyError as err:
            raise LogKeyError(
                f"Cannot append to unknown run {run!r}"
            ) from err
        record[key] = value
        self.dump(contents)
        return self

    def mv(self, source: paths.PathLike, target: paths.PathLike):
        """Rename `source` to `target` in this log file."""
        old = str(source)
        new = str(target)
        if old not in self:
            raise LogKeyError(
                f"Cannot rename unknown run {old!r}"
            ) from None
        updated = {k: v for k, v in self._asdict.items() if k != old}
        updated[new] = self._asdict[old]
        self.dump(updated)
        return self

    def rm(self, target: paths.PathLike):
        """Remove the target run from this log file."""
        updated = {k: v for k, v in self._asdict.items() if k != str(target)}
        self.dump(updated)
        return self

    @property
    def _asdict(self) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """Internal dictionary representing the current contents."""
        with self.path.open('r') as fp:
            return dict(json.load(fp))

    def dump(self, contents: typing.Mapping):
        """Write `contents` to this log file."""
        payload = {
            k: str(v) if isinstance(v, pathlib.Path) else v
            for k, v in contents.items()
        }
        with self.path.open('w') as fp:
            json.dump(payload, fp, indent=4, sort_keys=True)

    @property
    def name(self):
        """The name of this log file.
        
        Same as `RunLog.path.name`.
        """
        return self.path.name

    @property
    def path(self):
        """The path to this log file."""
        return self._path

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self._asdict)

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self.path})"


class ProjectExistsError(Exception):
    """A project with this name already exists."""


_Paths = typing.TypeVar('_Paths')
_Paths = typing.Union[pathlib.Path, typing.Tuple[pathlib.Path, pathlib.Path]]


class _PathOperation(collections.abc.Iterable):
    """Helper for path-based operations."""

    @typing.overload
    def __init__(
        self,
        operator: typing.Callable[
            [pathlib.Path],
            typing.Union[pathlib.Path, paths.PathOperationError]
        ],
        paths: typing.Iterable[pathlib.Path]=None,
    ) -> None: ...

    @typing.overload
    def __init__(
        self,
        operator: typing.Callable[
            [pathlib.Path, pathlib.Path],
            typing.Union[pathlib.Path, paths.PathOperationError]
        ],
        paths: typing.Iterable[typing.Tuple[pathlib.Path, pathlib.Path]]=None,
    ) -> None: ...

    def __init__(self, operator, paths=None) -> None:
        """Create a new instance."""
        self.operator = operator
        self._paths = paths
        self._force = None
        self._silent = None

    def __bool__(self) -> bool:
        """Called for bool(self)."""
        return bool(self._paths)

    def __iter__(self):
        """Iteratively apply the operation, if possible."""
        error = None
        for path in self._paths:
            result = self.apply(path)
            if not isinstance(result, paths.PathOperationError):
                yield result
                error = False
            else:
                error = True
            if error:
                if not self._force:
                    raise result
                if not self._silent:
                    print(result)

    def errors(self, force: bool=False, silent: bool=False):
        """Update how this instance handles errors and warnings."""
        self._force = force
        self._silent = silent
        return self

    def apply(
        self,
        this
    ) -> typing.Union[pathlib.Path, typing.Tuple[pathlib.Path]]:
        """Apply the operator to `this`, if possible."""
        if isinstance(this, pathlib.Path):
            return self.operator(this)
        try:
            allpaths = all(isinstance(i, pathlib.Path) for i in this)
        except TypeError:
            allpaths = False
        if allpaths:
            return self.operator(*this)


# TODO: Python >= 3.9
# - list[typing.Union[pathlib.Path, tuple[pathlib.Path, ...]]]

# TODO: Python >= 3.10
# - list[pathlib.Path | tuple[pathlib.Path, ...]]

ListOfPathsOrPathTuples = typing.TypeVar('ListOfPathsOrPathTuples')
ListOfPathsOrPathTuples = typing.List[
    typing.Union[pathlib.Path, typing.Tuple[pathlib.Path, ...]]
]

RunPathsType = typing.TypeVar('RunPathsType', bound='RunPaths')


class RunPaths(collections.abc.Collection):
    """Manager for path operations on runtime directories."""

    def __init__(
        self,
        root: paths.PathLike,
        branches: typing.Iterable[str]=None,
        base: str=None,
    ) -> None:
        self._root = paths.normalize(root)
        self._base = base or 'runs'
        self._branches = branches
        self._listing = None
        for directory in self.listing:
            directory.mkdir(parents=True, exist_ok=True)

    def update(
        self: RunPathsType,
        root: paths.PathLike=None,
        branches: typing.Iterable[str]=None,
        base: str=None,
    ) -> RunPathsType:
        """Update path components."""
        self._listing = None
        if root:
            path = paths.normalize(root)
            if path.exists():
                raise paths.PathTypeError(
                    f"Renaming {self.root.name!r} to {path.name!r} would "
                    f"overwrite {path}."
                )
            self.root.rename(path)
            self._root = path
        if branches:
            self._branches = branches
        if base:
            self._base = base
        return self

    def __contains__(self, __x: paths.PathLike) -> bool:
        """Called for __x in self."""
        return __x in self.listing

    def __iter__(self) -> typing.Iterator[pathlib.Path]:
        """Called for iter(self)."""
        return iter(self.listing)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.listing)

    def mkdir(
        self,
        target: paths.PathLike,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> _PathOperation:
        """Create target run(s) if possible."""
        return _PathOperation(
            self._mkdir,
            paths=self.define(target, branches=branches),
        )

    def _mkdir(self, target: paths.PathLike):
        """Create the target run if safe to do so."""
        this = paths.normalize(target)
        if this.exists():
            return paths.PathOperationError(
                f"Cannot create {str(this)!r}: already exists"
            )
        this.mkdir(parents=True)
        return this

    def mv(
        self,
        source: paths.PathLike,
        target: paths.PathLike,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> _PathOperation:
        """Rename `source` to `target` where possible."""
        pairs = self.define(source, target, branches=branches)
        return _PathOperation(self._mv, paths=pairs)

    def _mv(self, source: paths.PathLike, target: paths.PathLike):
        """Rename `source` to `target` only if safe to do so."""
        this = paths.normalize(source)
        that = paths.normalize(target)
        if not this.exists():
            return paths.PathOperationError(
                f"Cannot rename {this}: does not exist"
            )
        if not this.is_dir():
            return paths.PathOperationError(
                f"Cannot rename {this}: not a directory"
            )
        if that.exists():
            return paths.PathOperationError(
                f"Renaming {this.name!r} to {that.name!r} would "
                f"overwrite {that}."
            )
        this.rename(that)
        return (this, that)

    def rm(
        self,
        target: paths.PathLike,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> _PathOperation:
        """Remove target run(s), if possible."""
        return _PathOperation(
            self._rm,
            paths=self.define(pattern=target, branches=branches),
        )

    def _rm(self, target: paths.PathLike):
        """Remove the target path only if safe to do so."""
        this = paths.normalize(target)
        if not this.exists():
            return paths.PathOperationError(
                f"Cannot remove {this}: does not exist"
            )
        if not this.is_dir():
            return paths.PathOperationError(
                f"Cannot remove {this}: not a directory"
            )
        shutil.rmtree(this)
        return this

    def define(
        self,
        *names: str,
        pattern: str=None,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> ListOfPathsOrPathTuples:
        """Build a list of paths or tuples of paths.
        
        Parameters
        ----------
        *names : string
            Zero or more names of runtime directories to create.
        """
        rundirs = self._get_rundirs(branches)
        if pattern:
            return [
                path for rundir in rundirs
                for path in rundir.glob(pattern)
            ]
        if len(names) == 1:
            name = names[0]
            return [rundir / name for rundir in rundirs]
        if not names:
            return self.define(pattern='*', branches=branches)
        return [tuple(rundir / name for name in names) for rundir in rundirs]

    def resolve(self, target: str, branches: typing.Iterable[str]=None):
        """Compute the full path(s) to the target run."""
        if not branches:
            return [self.root / self.base / target]
        return [
            self.root / branch / self.base / target
            for branch in branches
        ]

    @property
    def runs(self) -> typing.Dict[str, typing.Set[str]]:
        """The available runs, and owning branches, if any."""
        if not self._branches:
            return {
                run.name: set()
                for run in (self.root / self.base).glob('*')
            }
        base = {
            run.name: set()
            for branch in self._branches
            for run in (self.root / branch / self.base).glob('*')
        }
        for run, branches in base.items():
            branches.update(
                branch for branch in self._branches
                if (self.root / branch / self.base / run).is_dir()
            )
        return base

    @property
    def branches(self):
        """The project branches, if any, and their available runs."""
        if not self._branches:
            return {}
        base = {branch: set() for branch in self._branches}
        for run, branches in self.runs.items():
            for branch in branches:
                base[branch].update({run})
        return base

    def _get_rundirs(
        self,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> typing.List[pathlib.Path]:
        """Build an appropriate collection of runtime directories."""
        if not branches:
            return self.listing
        subset = (
            {branches} if isinstance(branches, str)
            else set(branches)
        )
        return [d for d in self.listing if d.parent.name in subset]

    @property
    def listing(self):
        """Full paths to all runtime directories.
        
        Notes
        -----
        This property does not use the `RunPaths.branches` property to iterate
        over available branches because that would create an infinite recursion
        via `RunPaths._get_rundirs`.
        """
        if self._listing is None:
            self._listing = [
                self.root / branch / self.base
                for branch in self._branches
            ] if self._branches else [self.root / self.base]
        return self._listing

    @property
    def root(self):
        """The root directory."""
        return self._root

    @property
    def base(self):
        """The name of the common base runtime directory."""
        return self._base


class Attrs:
    """A container for `Interface` attributes."""

    def __init__(self, path: paths.PathLike, **kwargs) -> None:
        root = paths.normalize(path)
        branches = tuple(kwargs.get('branches') or ())
        config = str(kwargs.get('config') or 'eprem.cfg')
        inputs = paths.localize(kwargs.get('inputs') or 'inputs')
        output = paths.localize(kwargs.get('output') or 'eprem.log')
        rundir = paths.localize(kwargs.get('rundir') or 'runs')
        logname = paths.localize(
            kwargs.get('logname') or 'runs'
        ).with_suffix('.json')
        self._public = {
            'root': root,
            'branches': branches,
            'config': config,
            'inputs': root / inputs,
            'output': output,
            'rundir': rundir,
            'logname': logname,
        }
        self._serial = None
        self._options = None

    @property
    def options(self):
        """The values of optional attributes."""
        if self._options is None:
            self._options = {
                k: v for k, v in self._public.items()
                if k != 'root'
            }
        return self._options

    @property
    def serial(self):
        """A serializable representation of this object."""
        if self._serial is None:
            self._serial = {
                k: self._serialize(v)
                for k, v in self._public.items()
            }
        return self._serial

    def _serialize(self, this):
        """Convert `this` to a serializable type."""
        if isinstance(this, pathlib.Path):
            return str(this)
        return this

    @property
    def root(self) -> pathlib.Path:
        """The project root directory."""
        return self._public['root']

    @property
    def branches(self) -> typing.Tuple[str]:
        """The distinct project branches, if any."""
        return self._public['branches']

    @property
    def config(self) -> str:
        """The name of the standard project configuration file."""
        return self._public['config']

    @property
    def inputs(self) -> pathlib.Path:
        """The project subdirectory containing configuration files."""
        return self._public['inputs']

    @property
    def output(self) -> pathlib.Path:
        """The local path component of the project output log."""
        return self._public['output']

    @property
    def rundir(self) -> pathlib.Path:
        """The local path component of the project run directory."""
        return self._public['rundir']

    @property
    def logname(self) -> pathlib.Path:
        """The local path component of the project log file."""
        return self._public['logname']

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        display = {k: v for k, v in self._public.items() if k != 'branches'}
        display['branches'] = self._public['branches'] or None
        return '\n'.join(f"{k}: {v}" for k, v in display.items())


InterfaceType = typing.TypeVar('InterfaceType', bound='Interface')


class Interface:
    """Interface to a group of EPREM runs."""

    def __init__(
        self,
        root: paths.PathLike,
        branches: typing.Iterable[str]=None,
        config: str=None,
        inputs: paths.PathLike=None,
        output: str=None,
        rundir: str=None,
        logname: str=None,
        copy: paths.PathLike=None,
    ) -> None:
        """Initialize a new project."""
        self._isvalid = False
        attrs = self._init_project(
            root,
            copy=copy,
            branches=branches,
            config=config,
            inputs=inputs,
            output=output,
            rundir=rundir,
            logname=logname,
        )
        self._log = None
        self._name = None
        self._root = None
        self._log = self._get_log(attrs)
        self._directories = RunPaths(attrs.root, attrs.branches, attrs.rundir)
        self._setup_cli(attrs)
        attrs.inputs.mkdir(parents=True, exist_ok=True)
        if copy:
            copied = self._restore_project(paths.normalize(copy))
            shutil.copytree(copied.inputs, attrs.inputs, dirs_exist_ok=True)
        self._attrs = attrs
        self._isvalid = True

    def _init_project(
        self,
        root: paths.PathLike,
        copy: paths.PathLike=None,
        **options,
    ) -> Attrs:
        """Initialize arguments from input or the database."""
        if not copy:
            return self._get_attrs(root, **options)
        restored = self._restore_project(paths.normalize(copy))
        copied = restored.options
        copied.update(options)
        return self._get_attrs(root, **copied)

    def _get_attrs(self, root: paths.PathLike, **options):
        """Create or restore project attributes."""
        path = paths.normalize(root)
        # If the project doesn't exist, create it.
        if not path.exists():
            return self._create_project(path, **options)
        # If the project exists and the user isn't attempting to overwrite it,
        # reinitialize the interface.
        if not options or all(v is None for v in options.values()):
            return self._restore_project(path)
        existing = (
            f"{self.__class__.__qualname__}"
            f"({os.path.relpath(path)!r})"
        )
        # If the project exists and the user is attempting to overwrite it,
        # raise an error.
        raise ProjectExistsError(
            f"The project {path.name!r} already exists in {path.parent}. "
            f"You can access the existing project via {existing}"
        )

    def _create_project(self, path: pathlib.Path, **options):
        """Set attributes on a new project."""
        attrs = Attrs(path, **options)
        attrs.root.mkdir(parents=True)
        with (path / '.eprem.conf').open('w') as fp:
            json.dump(attrs.serial, fp, indent=4, sort_keys=True)
        return attrs

    def _restore_project(self, path: pathlib.Path):
        """Load attributes of an existing project."""
        with (path / '.eprem.conf').open('r') as fp:
            old = dict(json.load(fp))
        here = old.pop('root')
        return Attrs(here, **old)

    def _get_log(self, attrs: Attrs):
        """Create or retrieve the log of runs."""
        path = attrs.root / attrs.logname
        if path.exists():
            return RunLog(path)
        return RunLog(
            path,
            config=attrs.config,
            output=attrs.output,
        )

    def _setup_cli(self, attrs: Attrs):
        """Make sure this project can run as a script."""
        path = attrs.root / '__main__.py'
        if path.exists():
            return
        with (DIRECTORY / '_project_main.py').open('r') as rp:
            lines = rp.readlines()
        payload = lines.copy()
        import_line = lines.index(f'import {__name__}\n')
        path_spec = (
            "import pathlib\n"
            "import sys\n"
            "\n"
            f"sys.path.append({str(DIRECTORY)!r})\n"
            "\n"
        )
        payload.insert(import_line, path_spec)
        with path.open('w') as wp:
            wp.writelines(payload)
            wp.write("if __name__ == '__main__':\n")
            wp.write("    cli.run()\n")
            wp.write("\n")
        if _HAVE_ISORT:
            isort.file(path, quiet=True)
        sys.path.append(attrs.root)

    def show(self: InterfaceType, *runs: str):
        """Display information about this project or the named run(s).
        
        Parameters
        ----------
        *runs : string
            The named run(s) to display. This method also accepts `'*'`, which
            will cause it to display information about all available runs.
        """
        if not runs:
            self._show_project()
        requested = (
            tuple(self.runs) if len(runs) == 1 and runs[0] == '*'
            else runs
        )
        for run in requested:
            etc.underline(run)
            self._show_run(run)

    def _show_project(self):
        """Display information about this project."""
        etc.underline("Project")
        print(self._attrs)
        if not self.branches:
            etc.underline("Runs")
            print('\n'.join(self.runs))
            return
        for branch, runs in self.branches.items():
            etc.underline(f"Branch {branch}")
            print('\n'.join(runs))

    def _show_run(self, run: str):
        """Display information about the named run."""
        try:
            branches = self.runs[run]
        except KeyError:
            print(f"No run named {run!r}")
            return
        for path in self.directories.resolve(run, branches):
            print(path)

    def reset(self, force: bool=False, silent: bool=False):
        """Reset this project to its initial state."""
        return self.rm('*', errors=(not force), silent=silent)

    def rename(self, target: str, force: bool=False, silent: bool=False):
        """Rename this project to `target`."""
        # Save the current root directory since it will change.
        old = self.root
        # Convert the target to a full path.
        new = paths.normalize(target)
        # Handle an existing target path.
        if new.exists():
            if not force:
                raise paths.PathOperationError(
                    f"Renaming this project to {target!r} would "
                    f"overwrite {new}."
                )
            if not silent:
                print(f"Overwriting {new}")
        # Update the runtime paths on disk.
        self._directories.update(root=new)
        # Update the log of runtime paths.
        self.log.update_directory(new)
        # Update the project metadata.
        self._update_metadata(str(old), str(new), new / '.eprem.conf')
        # Echo success, if applicable.
        if not silent:
            if self.root.parent == old.parent == pathlib.Path.cwd():
                result = target
            else:
                result = str(self.root)
            print(f"Renamed project to {result!r}")

    def _update_metadata(
        self,
        source: str,
        target: str,
        metadata: pathlib.Path,
    ) -> None:
        """Update this project's path-related metadata."""
        with metadata.open('r') as fp:
            current = dict(json.load(fp))
        updated = {k: v for k, v in current.items() if k != source}
        updated['root'] = target
        with metadata.open('w') as fp:
            json.dump(updated, fp, indent=4, sort_keys=True)
        with contextlib.suppress(ValueError):
            sys.path.remove(source)
        sys.path.append(target)

    def remove(self, force: bool=False, silent: bool=False):
        """Delete this project."""
        shutil.rmtree(self.root, ignore_errors=force)
        with contextlib.suppress(ValueError):
            sys.path.remove(str(self.root))
        self._isvalid = False
        if not silent:
            print(f"Removed project at {self.root}")

    def run(
        self: InterfaceType,
        config: paths.PathLike,
        name: str=None,
        branches: typing.Union[str, typing.Iterable[str]]=None,
        nproc: int=None,
        environment: typing.Dict[str, str]=None,
        errors: bool=False,
        silent: bool=False,
    ) -> InterfaceType:
        """Set up and execute a new EPREM run within this project."""
        run = name or datetime.datetime.now().strftime('%Y-%m-%dT%H-%M-%S.%f')
        p = self.directories.mkdir(run, branches=branches)
        for path in p.errors(force=(not errors), silent=silent):
            self._create_run(
                config=config,
                path=path,
                nproc=nproc,
                environment=environment,
                silent=silent,
            )

    def _create_run(
        self: InterfaceType,
        config: paths.PathLike,
        path: pathlib.Path,
        nproc: int=None,
        environment: typing.Dict[str, str]=None,
        silent: bool=False,
    ) -> None:
        """Create a single run."""
        prjcfg = self.root / 'inputs' / config
        inputs = prjcfg if prjcfg.is_file() else paths.fullpath(config, strict=True)
        shutil.copy(inputs, path / self._attrs.config)
        branch = path.parent.parent
        mpirun = paths.locate('mpirun', branch, environment or {})
        eprem = paths.locate('eprem', branch, environment or {})
        command = (
            "nice -n 10 ionice -c 2 -n 3 "
            f"{mpirun} --mca btl_base_warn_component_unused 0 "
            f"-n {nproc or 1} {eprem} eprem.cfg"
        )
        output = path / self._attrs.output
        now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        with output.open('w') as stdout:
            process = subprocess.Popen(
                command,
                shell=True,
                cwd=path,
                stdout=stdout,
                stderr=subprocess.STDOUT,
            )
            if not silent:
                print(f"[{process.pid}]")
                print(f"Started {path.name!r} at {now}")
            process.wait()
            if not silent:
                base = f"Finished {path.name!r}"
                message = (
                    f"{base} in branch {branch.name!r}" if self.branches
                    else base
                )
                print(message)
        if process.returncode == 0:
            logentry = {
                'command': command,
                'time': now,
            }
            self.log.create(str(path), logentry)
        elif not silent:
            print(
                f"WARNING: Process exited with {process.returncode}",
                end='\n\n',
            )

    def mv(
        self: InterfaceType,
        source: str,
        target: str,
        branches: typing.Union[str, typing.Iterable[str]]=None,
        errors: bool=False,
        silent: bool=False,
    ) -> InterfaceType:
        """Rename an existing EPREM run within this project."""
        pairs = self.directories.mv(source, target, branches=branches)
        if not pairs:
            if not silent:
                print(f"Nothing to rename for {source!r}")
            return
        for (run, new) in pairs.errors(force=(not errors), silent=silent):
            self.log.mv(run, new)
            if not silent:
                branch = self._get_branch_name(run)
                base = f"Renamed {source!r} to {target!r}"
                print(f"{base} in branch {branch!r}" if branch else base)

    def rm(
        self: InterfaceType,
        run: str,
        branches: typing.Union[str, typing.Iterable[str]]=None,
        errors: bool=False,
        silent: bool=False,
    ) -> InterfaceType:
        """Remove an existing EPREM run from this project."""
        p = self.directories.rm(run, branches=branches)
        if not p:
            if not silent:
                print(f"Nothing to remove for {run!r}")
            return
        for path in p.errors(force=(not errors), silent=silent):
            self.log.rm(path)
            if not silent:
                base = f"Removed {path.name!r}"
                branch = self._get_branch_name(path)
                print(f"{base} from branch {branch!r}" if branch else base)

    def _get_branch_name(self, path: pathlib.Path):
        """Get the project branch name, if any, of `path`."""
        parents = (str(p) for p in path.relative_to(self.root).parents)
        with contextlib.suppress(StopIteration):
            if this := next(p for p in parents if p in self.branches):
                return this

    def _get_rundirs(
        self,
        branches: typing.Union[str, typing.Iterable[str]]=None,
    ) -> typing.List[pathlib.Path]:
        """Build an appropriate collection of runtime directories."""
        if not branches:
            return self.directories
        subset = (
            {branches} if isinstance(branches, str)
            else set(branches)
        )
        return [d for d in self.directories if d.parent.name in subset]

    @property
    def log(self):
        """The log of runs in this project."""
        return self._log

    @property
    def name(self):
        """The name of this project.
        
        This property is an alias for `~Interface.root.name`.
        """
        if self._name is None:
            self._name = self.root.name
        return self._name

    @property
    def runs(self):
        """The available runs, and owning branches, if any."""
        return self.directories.runs

    @property
    def branches(self):
        """The project branches, if any, and their available runs."""
        return self.directories.branches

    @property
    def root(self):
        """The top-level directory of this project."""
        return self.directories.root

    @property
    def directories(self):
        """The full path to each run directory."""
        return self._directories

    def __bool__(self) -> bool:
        """True if this is a valid project."""
        return self._isvalid

    def __eq__(self, other) -> bool:
        """True if two projects have the same initializing attributes."""
        if isinstance(other, Interface):
            return self._attrs == other._attrs
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.name

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self.root})"


cli = interfaces.CLI(
    formatter_class=argparse.RawTextHelpFormatter,
    description=etc.doc2help(__doc__),
)


@cli.include
def create(
    path: paths.PathLike,
    branches: typing.Union[str, typing.Iterable[str]]=None,
    config: str=None,
    inputs: str=None,
    output: str=None,
    rundir: str=None,
    logname: str=None,
    copy: paths.PathLike=None,
    verbose: bool=False,
) -> None:
    """Create a new project."""
    project = Interface(
        path,
        branches=branches,
        config=config,
        inputs=inputs,
        output=output,
        rundir=rundir,
        logname=logname,
        copy=copy,
    )
    if verbose:
        parts = [
            f"Created project {project.name!r} in",
            f"{project.root.parent}",
        ]
        single = ' '.join(parts)
        message = single if len(single) < 70 else '\n'.join(parts)
        print(message)
cli.subcommands['create'].add_argument(
    'path',
    help="the path to the target project; may be relative",
)
cli.subcommands['create'].add_argument(
    '-b',
    '--branches',
    help="names of project branches\n(default: none)",
    nargs='*',
    metavar=('A', 'B'),
)
cli.subcommands['create'].add_argument(
    '-c',
    '--config',
    help=(
        "the name to give runtime config files"
        "\n(default: 'eprem.cfg')"
    ),
)
cli.subcommands['create'].add_argument(
    '-i',
    '--inputs',
    help=(
        "the name of the project subdirectory that will contain"
        "\nruntime configuration files"
        "\n(default 'inputs')"
    )
)
cli.subcommands['create'].add_argument(
    '-o',
    '--output',
    help=(
        "the name to give runtime output logs"
        "\n(default: 'eprem.log')"
    ),
)
cli.subcommands['create'].add_argument(
    '-d',
    '--rundir',
    help=(
        "the name of the directory that will contain simulation runs"
        "\n(default: 'runs')"
    ),
)
cli.subcommands['create'].add_argument(
    '-l',
    '--logname',
    help=(
        "the name of the file to which to log run metadata"
        "\n(default: 'runs.json')"
    ),
)
cli.subcommands['create'].add_argument(
    '--copy',
    help="create a copy of PROJECT",
    metavar='PROJECT',
)
cli.subcommands['create'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def reset(
    path: paths.PathLike,
    force: bool=False,
    verbose: bool=False,
) -> None:
    """Reset an existing project."""
    project = Interface(path)
    project.reset(force=force, silent=(not verbose))
cli.subcommands['reset'].add_argument(
    'path',
    help="the path to the target project; may be relative",
)
cli.subcommands['reset'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when resetting a project",
    action='store_true',
)
cli.subcommands['reset'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def rename(
    path: paths.PathLike,
    target: paths.PathLike,
    force: bool=False,
    verbose: bool=False,
) -> None:
    """Rename an existing EPREM project."""
    project = Interface(path)
    project.rename(target, force=force, silent=(not verbose))
cli.subcommands['rename'].add_argument(
    'path',
    help="the path to the target project; may be relative",
)
cli.subcommands['rename'].add_argument(
    'target',
    help="the new name of the project",
)
cli.subcommands['rename'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when renaming a project",
    action='store_true',
)
cli.subcommands['rename'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def remove(
    path: paths.PathLike,
    force: bool=False,
    verbose: bool=False,
) -> None:
    """Remove an existing EPREM project."""
    project = Interface(path)
    project.remove(force=force, silent=(not verbose))
cli.subcommands['remove'].add_argument(
    'path',
    help="the path to the target project; may be relative",
)
cli.subcommands['remove'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when a removing project",
    action='store_true',
)
cli.subcommands['remove'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def show(
    path: paths.PathLike,
    runs: typing.Union[str, typing.Iterable[str]]=None,
) -> None:
    """Display information about an existing project."""
    project = Interface(path)
    project.show(*list(runs or []))
cli.subcommands['show'].add_argument(
    'path',
    help="the path to the target project; may be relative",
)
showruns = cli.subcommands['show'].add_mutually_exclusive_group()
showruns.add_argument(
    '-r',
    '--run',
    dest='runs',
    help=(
        "display information about the named run"
        ";\nspecify multiple times to include multiple runs"
    ),
    action='extend',
)
showruns.add_argument(
    '-a',
    '--all',
    dest='runs',
    help="display information for all runs",
    action='store_const',
    const='*',
)


if __name__ == '__main__':
    cli.run()




"""
Objects and utilities relevant to metric properties and systems.
"""

import typing

from ._conversions import (
    convert,
)
from ._defined import (
    UNITS,
    DIMENSIONS,
    NamedUnit,
    Reduction,
    identify,
    ratio,
)
from ._dimensions import (
    dimension_factory as dimension,
    Dimension,
)
from ._exceptions import (
    UnitConversionError,
    UnitParsingError,
)
from ._reference import (
    CONVERSIONS,
    SYSTEMS,
    standardize,
)
from ._systems import (
    system_factory as system,
    quantity_factory as quantity,
    Quantity,
    System,
)
from ._types import (
    Interface,
    Type,
)
from ._units import (
    normalize,
    unit_factory as unit,
    Unit,
    UnitLike,
    unitlike,
)

__all__ = [
    'CONVERSIONS',
    'DIMENSIONS',
    'SYSTEMS',
    'UNITS',
    'Conversion',
    'Dimension',
    'Interface',
    'NamedUnit',
    'Reduction',
    'Quantity',
    'System',
    'Type',
    'Unit',
    'UnitConversionError',
    'UnitLike',
    'UnitParsingError',
    'convert',
    'dimension',
    'identify',
    'normalize',
    'quantity',
    'ratio',
    'standardize',
    'system',
    'unit',
    'unitlike',
]


class Conversion(_conversions.Conversion):
    """The result of a unit conversion."""

    @property
    def old(self):
        """The original unit."""
        return unit(self.u0)

    @property
    def new(self):
        """The converted unit."""
        return unit(self.u1)


@typing.overload
def conversion(
    source: typing.Union[str, Unit],
    target: typing.Union[str, Unit],
    quantity: typing.Optional[str]=None,
) -> Conversion: ...

@typing.overload
def conversion(
    source: typing.Union[str, Unit],
    target: typing.Union[str, Unit],
    error: typing.Literal[False],
    quantity: typing.Optional[str]=None,
) -> typing.Optional[Conversion]: ...

def conversion(source, target, **kwargs):
    """Compute the conversion from `source` to `target`, if possible.

    Parameters
    ----------
    source : string or `~metric.Unit`
        The unit from which to convert.

    target : string or `~metric.Unit`
        The unit or metric system to which to convert.

    error : bool, default=True
        If false, return `None` for a failed conversion. The default behavior is
        to allow undefined conversions to raise `~metric.UnitConversionError`.

    quantity : string, optional
        The physical quantity associated with `source`.

    Returns
    -------
    `~Conversion` or `None`
        An object representing the conversion from `source` to `target`,
        including the relevant numerical factor, if the conversion was
        successful. The value of `error` determines the behavior if the
        conversion fails.

    Examples
    --------
    As a simple first example, suppose you have something that is 2 meters long
    and you want to convert its length to centimeters:

    >>> 2 * float(metric.conversion('m', 'cm'))
    200.0

    This simple case is equivalent to computing the metric ratio of meter to
    centimeter because they have the same base unit:

    >>> 2 * metric.ratio('m', 'cm')
    200.0

    As a more complex example, suppose you have an amount of energy flux in 'J /
    au^2'. In order to convert to the equivalent amount in 'eV / km^2', you must
    multiply by

    >>> float(metric.conversion('J / au^2', 'eV / km^2'))
    278.8896829059863

    which is the same as first multiplying by the factor to convert from 'J' to
    'eV', then dividing by the square of the factor to convert from 'au' to
    'km', as shown below

    >>> c0 = metric.conversion('J', 'eV')
    >>> c1 = metric.conversion('au', 'km')
    >>> float(c0)
    6.241418050181001e+18
    >>> float(c1)
    149597870.70000002
    >>> float(c0) / float(c1)**2
    278.8896829059863

    This function also supports converting the original unit to the equivalent
    unit in a given metric system

    >>> metric.conversion('G', 'mks')
    Conversion('G', 'T', 0.0001)
    """
    try:
        result = convert(
            str(source),
            str(target),
            quantity=kwargs.get('quantity'),
        )
    except _exceptions.UnitConversionError as err:
        if kwargs.get('error', True):
            raise err
        return
    return Conversion(result.u0, result.u1, float(result))





"""
Support for numeric objects with real-valued data.
"""

import numbers
import typing

import numpy.typing

from .. import base
from .. import container
from ._types import ValueType
from ._objects import (
    Variable,
)
from ._array import (
    Array,
    array_factory as array,
)


__all__ = [
    'Array',
    'ValueType',
    'Variable',
    'array',
]


ArrayLike = typing.Union[
    numpy.typing.NDArray[numpy.integer],
    numpy.typing.NDArray[numpy.floating],
    typing.Sequence[numbers.Real],
]
"""Type alias for array-like objects with real-valued data."""
# TODO: Include nested sequences.


def variable(
    x: typing.Union[ValueType, typing.Sequence[ValueType]],
    /,
) -> Variable[ValueType]:
    """Create a real-valued variable.
    
    Parameters
    ----------
    x : real number or sequence of real numbers
        The real-valued object from which to create a new variable.

    Returns
    -------
    `~Variable`

    Raises
    ------
    `TypeError`
        The caller passed a single object of non-numeric type or a sequence that
        includes members of non-numeric type.
    """
    if container.isseparable(x):
        if all(base.isreal(v) for v in x):
            return Variable(x)
        raise TypeError(
            f"Cannot create a variable from {x!r}"
            " with non-numeric values."
        )
    if not base.isreal(x):
        raise TypeError(
            f"Cannot create variable from object of type {type(x)}"
        ) from None
    return Variable([x])


class Template(base.abc.Real):
    """Abstract class template for a real-valued object.
    
    The body of this class definition simply defines each real-valued operator
    (i.e., an operator required by the real-valued protocol) by reference to the
    corresponding abstract method on `~base.abc.Real`.This class is therefore
    isomorphic to `~base.abc.Real`. It exists to serve as a template for copying
    method placeholders to a concrete implementation of the real-valued
    protocol.

    See Also
    --------
    `~base.Real`
        Definition of the real-valued protocol.
    `~base.abc.Real`
        Abstract base class corresponding to `~base.Real`.
    """

    def __eq__(self, other):
        return super().__eq__(other)

    def __ne__(self, other):
        return super().__ne__(other)

    def __lt__(self, other):
        return super().__lt__(other)

    def __le__(self, other):
        return super().__le__(other)

    def __gt__(self, other):
        return super().__gt__(other)

    def __ge__(self, other):
        return super().__ge__(other)

    def __abs__(self):
        return super().__abs__()

    def __pos__(self):
        return super().__pos__()

    def __neg__(self):
        return super().__neg__()

    def __add__(self, other):
        return super().__add__(other)

    def __radd__(self, other):
        return super().__radd__(other)

    def __sub__(self, other):
        return super().__sub__(other)

    def __rsub__(self, other):
        return super().__rsub__(other)

    def __mul__(self, other):
        return super().__mul__(other)

    def __rmul__(self, other):
        return super().__rmul__(other)

    def __truediv__(self, other):
        return super().__truediv__(other)

    def __rtruediv__(self, other):
        return super().__rtruediv__(other)

    def __floordiv__(self, other):
        return super().__floordiv__(other)

    def __rfloordiv__(self, other):
        return super().__rfloordiv__(other)

    def __mod__(self, other):
        return super().__mod__(other)

    def __rmod__(self, other):
        return super().__rmod__(other)

    def __pow__(self, other):
        return super().__pow__(other)

    def __rpow__(self, other):
        return super().__rpow__(other)




import abc
import typing

from .. import base
from .. import measured
from .. import numeric
from .. import quantity
from ..typehelp import Self
from ._types import QuantityType as Q


class Operators(numeric.Operators, abc.ABC):
    """Abstract mixin class for custom physical operations."""

    @abc.abstractmethod
    def __eq__(self, other):
        """Called for self == other."""

    @abc.abstractmethod
    def __ne__(self, other):
        """Called for self != other."""

    @abc.abstractmethod
    def __lt__(self: Self, other) -> typing.Any:
        """Called for self < other."""

    @abc.abstractmethod
    def __gt__(self: Self, other) -> typing.Any:
        """Called for self > other."""

    @abc.abstractmethod
    def __le__(self: Self, other) -> typing.Any:
        """Called for self <= other."""

    @abc.abstractmethod
    def __ge__(self: Self, other) -> typing.Any:
        """Called for self >= other."""

    @abc.abstractmethod
    def __abs__(self: Self) -> Self:
        """Called for abs(self: Self)."""

    @abc.abstractmethod
    def __pos__(self: Self) -> Self:
        """Called for +self."""

    @abc.abstractmethod
    def __neg__(self: Self) -> Self:
        """Called for -self."""

    @abc.abstractmethod
    def __add__(self: Self, other) -> Self:
        """Called for self + other."""

    @abc.abstractmethod
    def __radd__(self: Self, other) -> Self:
        """Called for other + self."""

    @abc.abstractmethod
    def __sub__(self: Self, other) -> Self:
        """Called for self - other."""

    @abc.abstractmethod
    def __rsub__(self: Self, other) -> Self:
        """Called for other - self."""

    @abc.abstractmethod
    def __mul__(self: Self, other) -> Self:
        """Called for self * other."""

    @abc.abstractmethod
    def __rmul__(self: Self, other) -> Self:
        """Called for other * self."""

    @abc.abstractmethod
    def __truediv__(self: Self, other) -> Self:
        """Called for self / other."""

    @abc.abstractmethod
    def __rtruediv__(self: Self, other) -> Self:
        """Called for other / self."""

    @abc.abstractmethod
    def __floordiv__(self: Self, other) -> Self:
        """Called for self // other."""

    @abc.abstractmethod
    def __rfloordiv__(self: Self, other) -> Self:
        """Called for other // self."""

    @abc.abstractmethod
    def __mod__(self: Self, other) -> Self:
        """Called for self % other."""

    @abc.abstractmethod
    def __rmod__(self: Self, other) -> Self:
        """Called for other % self."""

    @abc.abstractmethod
    def __pow__(self: Self, other) -> Self:
        """Called for self ** other."""

    @abc.abstractmethod
    def __rpow__(self: Self, other) -> Self:
        """Called for other ** self."""


class Object(measured.Object[base.RealType], Operators):
    """A real-valued measurable object."""

    def __init__(
        self,
        data: base.RealType,
        context: measured.Context[Q],
        /,
    ) -> None:
        super().__init__(data, context)

    def __measure__(self) -> quantity.Measurement:
        return quantity.measurement(self.data, unit=self.unit)

    __eq__ = numeric.implementation.binary('eq')
    __ne__ = numeric.implementation.binary('ne')

    __lt__ = numeric.implementation.binary('lt')
    __gt__ = numeric.implementation.binary('gt')
    __le__ = numeric.implementation.binary('le')
    __ge__ = numeric.implementation.binary('ge')

    __abs__ = numeric.implementation.unary('abs')
    __pos__ = numeric.implementation.unary('pos')
    __neg__ = numeric.implementation.unary('neg')

    __add__, __radd__ = numeric.implementation.doublet('add')
    __sub__, __rsub__ = numeric.implementation.doublet('sub')
    __mul__, __rmul__ = numeric.implementation.doublet('mul')
    __truediv__, __rtruediv__ = numeric.implementation.doublet('truediv')
    __floordiv__, __rfloordiv__ = numeric.implementation.doublet('floordiv')
    __mod__, __rmod__ = numeric.implementation.doublet('mod')
    __pow__, __rpow__ = numeric.implementation.doublet('pow')


ObjectT = typing.TypeVar('ObjectT', bound=Object)
"""Type variable bound to `~physical.Object`."""





"""
Support for real-valued numeric arrays.
"""

import numbers
import typing

import numpy
import numpy.typing

from .. import base
from .. import numeric
from ._types import ValueType


T = typing.TypeVar('T')


class Array(numeric.Array[T], base.abc.Real):
    """A concrete array-like object with named dimensions.

    This class provides a real-valued array-like interface to a variety of
    numeric objects by implementing the real-valued protocol for a generic
    array-like quantity.

    See Also
    --------
    `~array`
        The factory function for creating instances of this class.
    `~numeric.Array`
        The base class for generic array-like quantities.
    """

    __eq__ = numeric.implementation.binary('eq')
    __ne__ = numeric.implementation.binary('ne')
    __lt__ = numeric.implementation.binary('lt')
    __gt__ = numeric.implementation.binary('gt')
    __le__ = numeric.implementation.binary('le')
    __ge__ = numeric.implementation.binary('ge')

    __abs__ = numeric.implementation.unary('abs')
    __pos__ = numeric.implementation.unary('pos')
    __neg__ = numeric.implementation.unary('neg')

    __add__, __radd__ = numeric.implementation.doublet('add')
    __sub__, __rsub__ = numeric.implementation.doublet('sub')
    __mul__, __rmul__ = numeric.implementation.doublet('mul')
    __truediv__, __rtruediv__ = numeric.implementation.doublet('truediv')
    __floordiv__, __rfloordiv__ = numeric.implementation.doublet('floordiv')
    __mod__, __rmod__ = numeric.implementation.doublet('mod')
    __pow__, __rpow__ = numeric.implementation.doublet('pow')

    def _get_arg_data(self, arg):
        if isinstance(arg, Array):
            return self.array
        return super()._get_arg_data(arg)

    _FUNCTIONS = {}
    _OPERATORS = {}


Array._UFUNC_TYPES += [Array]
Array._FUNCTION_TYPES += [Array]


ArrayType = typing.TypeVar('ArrayType', bound=Array)


@Array.implements('eq')
def eq(a, b):
    """Compute a == b."""
    if isinstance(a, Array) and isinstance(b, Array):
        return (
            numpy.array_equal(a, b)
            and
            a.dimensions == b.dimensions
        )
    return False


@Array.implements('ne')
def ne(a, b):
    """Compute a != b."""
    return not (a == b)


@Array.implements('abs')
def abs(a):
    """Compute abs(a)."""
    return unary(numeric.operators.abs, a)


@Array.implements('pos')
def pos(a):
    """Compute +a."""
    return unary(numeric.operators.pos, a)


@Array.implements('neg')
def neg(a):
    """Compute -a."""
    return unary(numeric.operators.neg, a)


@Array.implements('round')
def round(a, /, ndigits: int=None):
    """Compute round(a[, ndigits])."""
    return unary(numeric.operators.round, a, ndigits=ndigits)


@Array.implements('lt')
def lt(a, b):
    """Compute a < b."""
    return comparative(numeric.operators.lt, a, b)


@Array.implements('le')
def le(a, b):
    """Compute a <= b."""
    return comparative(numeric.operators.le, a, b)


@Array.implements('gt')
def gt(a, b):
    """Compute a > b."""
    return comparative(numeric.operators.gt, a, b)


@Array.implements('ge')
def ge(a, b):
    """Compute a >= b."""
    return comparative(numeric.operators.ge, a, b)


@Array.implements('add')
def add(a, b):
    """Compute a + b."""
    try:
        result = additive(numeric.operators.add, a, b)
    except ValueError as err:
        raise ValueError("Cannot add arrays") from err
    return result


@Array.implements('sub')
def sub(a, b):
    """Compute a - b."""
    try:
        result = additive(numeric.operators.sub, a, b)
    except ValueError as err:
        raise ValueError("Cannot subtract arrays") from err
    return result


@Array.implements(numpy.multiply)
@Array.implements('mul')
def mul(a, b):
    """Compute a * b."""
    return multiplicative(numeric.operators.mul, a, b)


@Array.implements(numpy.true_divide)
@Array.implements('truediv')
def truediv(a, b):
    """Compute a / b."""
    return multiplicative(numeric.operators.truediv, a, b)


@Array.implements(numpy.floor_divide)
@Array.implements('floordiv')
def floordiv(a, b):
    """Compute a // b."""
    return multiplicative(numeric.operators.floordiv, a, b)


@Array.implements(numpy.mod)
@Array.implements('mod')
def mod(a, b):
    """Compute a % b."""
    return multiplicative(numeric.operators.mod, a, b)


@Array.implements(numpy.power)
@Array.implements('pow')
def pow(a, b, mod: int=None):
    """Compute a ** b."""
    if isinstance(a, Array) and isinstance(b, Array):
        if a.dimensions != b.dimensions:
            raise ValueError(
                "Cannot compute a ** b between two arrays"
                " with different dimensions"
            ) from None
    f = numeric.operators.pow
    if isinstance(a, (numbers.Number, numpy.ndarray)):
        return f(a, b.array)
    return multiplicative(f, a, b, mod=mod)


def unary(f, a, /):
    """Implement a unary arithmetic operation on array quantities."""
    if isinstance(a, Array):
        return array_factory(f(a.array), dimensions=a.dimensions)
    raise numeric.OperandTypeError(a)


def comparative(f, a, b, /):
    """Implement a comparative operation on array quantities."""
    if isinstance(a, Array) and isinstance(b, Array):
        if not a.dimensions == b.dimensions:
            raise ValueError(
                "Cannot compare arrays with different dimensions"
            ) from None
        return f(a.array, b.array)
    return f(a.array, b)


def additive(f, a, b, /):
    """Implement an additive operation on array quantities."""
    if isinstance(a, Array) and isinstance(b, Array):
        if not a.dimensions == b.dimensions:
            raise ValueError("Arrays have different dimensions") from None
        return array_factory(f(a.array, b.array), dimensions=a.dimensions)
    if isinstance(a, Array):
        return f(a.array, b)
    if isinstance(b, Array):
        return f(a, b.array)
    raise numeric.OperandTypeError(a, b)


def multiplicative(f, a, b, /, **kwargs):
    """Implement a multiplicative operation on array quantities."""
    try:
        operands = _get_operands(a, b)
    except TypeError as err:
        raise numeric.OperandTypeError(a, b) from err
    r = f(*operands, **kwargs)
    if isinstance(a, Array) and isinstance(b, Array):
        return array_factory(r, a.dimensions | b.dimensions)
    if isinstance(a, Array):
        return array_factory(r, a.dimensions)
    if isinstance(b, Array):
        return array_factory(r, b.dimensions)
    raise numeric.OperandTypeError(a, b)


def _get_operands(a, b):
    """Compute appropriate operands for a binary operation."""
    if isinstance(a, Array) and isinstance(b, Array):
        return numeric.remesh(a, b)
    return _get_operand(a), _get_operand(b)


def _get_operand(x):
    """Determine the appropriate operand for a mixed-type operation.
    
    This is a helper function for `_get_operands`. It assumes that
    `_get_operands` has already handled the case in which both arguments are
    data interfaces.
    """
    if isinstance(x, Array):
        return x.array
    if isinstance(x, numeric.Object):
        return x.data
    if isinstance(x, (numbers.Real, numpy.ndarray)):
        return x
    raise TypeError(x)


@Array.implements(numpy.array_equal)
def array_equal(
    a: Array[ValueType],
    b: Array[ValueType],
) -> bool:
    """Called for numpy.array_equal(a, b)."""
    return numpy.array_equal(numpy.array(a), numpy.array(b))


@Array.implements(numpy.squeeze)
def squeeze(x: Array[ValueType], **kwargs):
    """Called for numpy.squeeze(x)."""
    data = numpy.squeeze(x.array, **kwargs)
    if data.ndim == 0:
        return float(data)
    dimensions = [d for d, n in x.shapemap.items() if n != 1]
    return array_factory(data, dimensions=dimensions)


@Array.implements(numpy.mean)
def mean(x: Array[ValueType], **kwargs):
    """Compute the mean of the underlying array."""
    data = x.array.mean(**kwargs)
    axis = kwargs.get('axis')
    if axis is None:
        return data
    ax = axis % x.ndim
    dimensions = [d for d in x.dimensions if x.dimensions.index(d) != ax]
    return array_factory(data, dimensions=dimensions)


@Array.implements(numpy.sum)
def sum(x: Array[ValueType], **kwargs):
    """Compute the sum of the underlying array."""
    data = x.array.sum(**kwargs)
    axis = kwargs.get('axis')
    if axis is None:
        return data
    if 'keepdims' in kwargs:
        return array_factory(data, dimensions=x.dimensions)
    ax = x.ndim + axis if axis < 0 else axis
    dimensions = [d for d in x.dimensions if x.dimensions.index(d) != ax]
    return array_factory(data, dimensions=dimensions)


@Array.implements(numpy.cumsum)
def cumsum(x: Array[ValueType], **kwargs):
    """Compute the cumulative sum of the underlying array."""
    data = x.array.cumsum(**kwargs)
    if kwargs.get('axis') is None:
        return data
    return array_factory(data, dimensions=x.dimensions)


@Array.implements(numpy.transpose)
def transpose(x: Array[ValueType], **kwargs):
    """Compute the transpose of an array and dimensions."""
    data = x.array.transpose()
    axes = kwargs.get('axes')
    if axes is None:
        dimensions = x.dimensions[::-1]
        return array_factory(data, dimensions=dimensions)
    dimensions = [x.dimensions[i] for i in tuple(axes)]
    return array_factory(data, dimensions=dimensions)


@Array.implements(numpy.gradient)
def gradient(x: Array[ValueType], *args, **kwargs):
    """Compute the gradient of an array."""
    if not args:
        return _apply_gradient(x.dimensions, x.array, **kwargs)
    diffs = []
    for arg in args:
        if isinstance(arg, Array):
            diffs.append(arg.array)
        elif isinstance(arg, numbers.Real):
            diffs.append(float(arg))
        else:
            diffs.append(numpy.array(arg))
    return _apply_gradient(x.dimensions, x.array, *diffs, **kwargs)


def _apply_gradient(
    dimensions,
    *args,
    **kwargs
) -> typing.List[Array]:
    """Helper for `gradient`."""
    gradient = numpy.gradient(*args, **kwargs)
    if isinstance(gradient, (list, tuple)):
        return [
            array_factory(array, dimensions=dimensions)
            for array in gradient
        ]
    return array_factory(gradient, dimensions=dimensions)


try:
    numpy.trapezoid
except AttributeError:
    numpy_trapezoid = numpy.trapz
else:
    numpy_trapezoid = numpy.trapezoid
@Array.implements(numpy_trapezoid)
def trapezoid(x: Array[ValueType], *args, **kwargs):
    """Integrate the array via the composite trapezoidal rule."""
    data = numpy_trapezoid(x.array, *args, **kwargs)
    axis = kwargs.get('axis')
    if axis is None:
        dimensions = x.dimensions[:-1]
    else:
        ax = axis % x.ndim
        dimensions = [d for d in x.dimensions if x.dimensions.index(d) != ax]
    return array_factory(data, dimensions=dimensions)


@typing.overload
def array_factory(*args, **kwargs) -> Array[T]:
    """Create an array from the given arguments, if possible."""

@typing.overload
def array_factory(x: T, /) -> Array[T]:
    """Create an array from `x`, with dimensions derived from `x`."""

@typing.overload
def array_factory(x: T, /, dimensions: typing.Iterable[str]) -> Array[T]:
    """Create an array from `x`, with the given dimensions."""

@typing.overload
def array_factory(x: T, *dimensions: str) -> Array[T]:
    """Create an array from `x`, with the given dimensions."""

@typing.overload
def array_factory(x: ArrayType, /) -> ArrayType:
    """Create an array from an existing array-like quantity."""

def array_factory(*args, **kwargs):
    return Array(*numeric._array.array_args(*args, **kwargs))





"""
Metadata for EPREM objects.
"""

import typing

from . import aliased
from . import datafile
from . import metric


_metadata = {
    'phiOffset': {
        'quantity': 'plane angle',
        'aliases': [],
    },
    'time': {
        'quantity': 'time',
        'aliases': ['t', 'times'],
    },
    'shell': {
        'quantity': 'number',
        'aliases': ['shells'],
    },
    'mu': {
        'quantity': 'ratio',
        'aliases': [
            'mu',
            'pitch angle', 'pitch-angle', 'pitch-angle cosine',
            'pitch angles', 'pitch-angles', 'pitch-angle cosines',
        ],
    },
    'mass': {
        'quantity': 'mass',
        'aliases': ['m'],
    },
    'charge': {
        'quantity': 'charge',
        'aliases': ['q'],
    },
    'energy': {
        'quantity': 'energy',
        'aliases': ['energies', 'E', 'egrid'],
    },
    'v': {
        'quantity': 'velocity',
        'aliases': ['speed', 'vgrid'],
    },
    'r': {
        'quantity': 'length',
        'aliases': ['radius', 'R'],
    },
    'theta': {
        'quantity': 'plane angle',
        'aliases': ['T'],
    },
    'phi': {
        'quantity': 'plane angle',
        'aliases': ['P'],
    },
    'br': {
        'quantity': 'magnetic field',
        'aliases': ['Br'],
    },
    'btheta': {
        'quantity': 'magnetic field',
        'aliases': ['bt', 'Btheta', 'Bt'],
    },
    'bphi': {
        'quantity': 'magnetic field',
        'aliases': ['bp', 'Bphi', 'Bp'],
    },
    'ur': {
        'quantity': 'velocity',
        'aliases': ['Ur', 'Vr', 'vr'],
    },
    'utheta': {
        'quantity': 'velocity',
        'aliases': ['ut', 'Utheta', 'Ut', 'Vtheta', 'Vt'],
    },
    'uphi': {
        'quantity': 'velocity',
        'aliases': ['up', 'Uphi', 'Up', 'Vphi', 'Vp'],
    },
    'rho': {
        'quantity': 'number density',
        'aliases': ['Rho'],
    },
    'f': {
        'quantity': 'particle distribution',
        'aliases': ['dist', 'Dist'],
    },
    'flux': {
        'quantity': 'number / (area * solid_angle * time * energy)',
        'aliases': ['Flux', 'J', 'J(E)', 'j', 'j(E)'],
    },
    'x': {
        'quantity': 'length',
        'aliases': ['X'],
    },
    'y': {
        'quantity': 'length',
        'aliases': ['Y'],
    },
    'z': {
        'quantity': 'length',
        'aliases': ['Z'],
    },
    'bmag': {
        'quantity': 'magnetic field',
        'aliases': ['B', 'b', '|B|', '|b|', 'b_mag', 'b mag'],
    },
    'umag': {
        'quantity': 'velocity',
        'aliases': ['U', 'u', '|U|', '|u|', 'u_mag', 'u mag'],
    },
    'upara': {
        'quantity': 'velocity',
        'aliases': ['Upara', 'u_para'],
    },
    'uperp': {
        'quantity': 'velocity',
        'aliases': ['Uperp', 'u_perp'],
    },
    'angle': {
        'quantity': 'plane angle',
        'aliases': ['flow_angle', 'flow angle'],
    },
    'divu': {
        'quantity': '1 / time',
        'aliases': ['div_u', 'divU', 'div U', 'div u', 'div(U)', 'div(u)'],
    },
    'rigidity': {
        'quantity': 'momentum / charge',
        'aliases': ['Rg', 'R_g'],
    },
    'mfp': {
        'quantity': 'length',
        'aliases': ['mean_free_path', 'mean free path'],
    },
    'ar': {
        'quantity': '1 / time',
        'aliases': ['acceleration_rate', 'acceleration rate'],
    },
    'energy_density': {
        'quantity': 'energy / volume',
        'aliases': ['energy density'],
    },
    'average_energy': {
        'quantity': 'energy',
        'aliases': ['average energy'],
    },
    'fluence': {
        'quantity': 'number / (area * solid_angle * energy / mass_number)',
        'aliases': [],
    },
    'intflux': {
        'quantity': 'number / (area * solid_angle * time)',
        'aliases': ['integral_flux', 'integral flux'],
    },
}


_quantities = {
    (k, *v.get('aliases', [])): v['quantity']
    for k, v in _metadata.items()
}
QUANTITIES = aliased.Mapping(_quantities)
"""An aliased mapping from simulated quantity to physical quantity."""


ALIASES = aliased.Groups(*QUANTITIES.keys(aliased=True))
"""The defined alias(es) for each simulated quantity."""


class Subset(typing.NamedTuple):
    """Reference information for a subset of EPREM quantities."""
    aliases: aliased.Groups
    """The defined alias(es) for each member."""
    names: aliased.NameMap
    """A mapping from alias(es) to canonical name for each member."""


_SUBSETS = {
    'arrays': datafile.ARRAYS,
    'scalars': datafile.SCALARS,
    'axes': datafile.AXES,
    'observables': tuple(_metadata),
}


def build(*names, subset: str=None):
    """Build reference information for EPREM quantities."""
    keys = _SUBSETS.get(subset, names)
    aliases = aliased.Groups(*[ALIASES.get(key, [key]) for key in keys])
    namemap = aliased.NameMap({key: ALIASES.get(key, [key]) for key in keys})
    return Subset(aliases=aliases, names=namemap)


ARRAYS = build(subset='arrays')
"""Reference information for EPREM arrays."""


SCALARS = build(subset='scalars')
"""Reference information for EPREM scalars."""


AXES = build(subset='axes')
"""Reference information for EPREM axes."""


OBSERVABLES = build(subset='observables')
"""Reference information for EPREM observables."""


_units = {
    k: metric.UNITS[q]
    for k, q in QUANTITIES.items(aliased=True)
}

UNITS = aliased.Mapping(_units)
"""The unit of each observable quantity.

This mapping is keyed by quantity alias, then by metric system.

See Also
--------
`~metric.UNITS`
"""


_dimensions = {
    k: metric.DIMENSIONS[q]
    for k, q in QUANTITIES.items(aliased=True)
}

DIMENSIONS = aliased.Mapping(_dimensions)
"""The dimension of each observable quantity.

This mapping is keyed by quantity alias, then by metric system.

See Also
--------
`~metric.UNITS`
"""




"""
Additional support for type annotations.

This module provides a single interface to extended type annotations. For
example, suppose `BestType` is available in the `typing` module starting with
Python version 3.X, and is available in the `typing_extensions` module for
earlier versions. This module will check the version of the active Python
interpreter and will import from either `typing` or `typing_extensions` as
appropriate, so that other modules may simply access `typehelp.BestType`.
"""

import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


__all__ = [
    "Self",
]




import typing

from ._operations import resolve


T = typing.TypeVar('T')


class Result:
    """Container class for the result of an operation."""

    def __init__(self, *args, **kwargs) -> None:
        self._args = args
        self._kwargs = kwargs

    def apply(self, f: typing.Callable[..., T]) -> T:
        """Create a new object by applying `f` to this result."""
        return f(*self.args, **self.kwargs)

    @property
    def args(self):
        """The declared positional arguments of this result."""
        return self._args

    @property
    def kwargs(self):
        """The declared keyword arguments of this result."""
        return self._kwargs


@typing.runtime_checkable
class Customizable(typing.Protocol):
    """Structural protocol for objects that implement custom operators."""
    def __object_operator__(self, f, *args, **kwargs) -> typing.Any: ...


def unary(operation: typing.Union[str, typing.Callable]):
    """Implement a unary numeric operation."""
    f = resolve(operation)
    def operator(self, **kwargs):
        if isinstance(self, Customizable):
            return self.__object_operator__(f, self, **kwargs)
        return NotImplemented
    return operator


def binary(operation: typing.Union[str, typing.Callable]):
    """Implement a binary numeric operation."""
    f = resolve(operation)
    def operator(self, other, **kwargs):
        if isinstance(self, Customizable):
            return self.__object_operator__(f, self, other, **kwargs)
        return NotImplemented
    return operator


def doublet(operation: typing.Union[str, typing.Callable]):
    """Implement a forward/reverse pair of binary numeric operations."""
    f = resolve(operation)
    def forward(self, other, **kwargs):
        if isinstance(self, Customizable):
            return self.__object_operator__(f, self, other, **kwargs)
        return NotImplemented
    def reverse(self, other, **kwargs):
        if isinstance(self, Customizable):
            return self.__object_operator__(f, other, self, **kwargs)
        return NotImplemented
    return forward, reverse





import typing

import numpy

from .. import base
from .. import container
from .. import real
from ._context import Context
from ._object import (
    Object,
    ObjectT,
)
from ._value import Value


class Sequence(Object[real.ValueType], base.mixins.Sequence):
    """A measured object with multiple values.

    See Also
    --------
    `~Object`
        An object with real-valued data and an associated unit.
    `~base.Sequence`
        The abstract protocol for sequence-like objects.
    `~base.mixins.Sequence`
        A partial implementation of the sequence-like protocol.
    """

    def __init__(
        self,
        data: typing.Sequence[real.ValueType],
        context: Context[ObjectT],
        /,
    ) -> None:
        super().__init__(data, context)
        self._size = None

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> typing.Iterator[Value]:
        return super().__iter__()

    def __getitem__(self, i: typing.SupportsIndex):
        data = self.data[i]
        # NOTE: A measured sequence should return a single measured value when
        # the subscripted data represents a single value. This is consistent
        # with the behavior of built-in sequences and numpy arrays. We approach
        # this problem by naively trying to create a measured value from the
        # subscripted data and returning that value if we succeeded. If anything
        # goes wrong when trying to create the measured value, we assume that
        # the subscripted data does not represent a single value, and return a
        # new measured sequence.
        try:
            result = Value.spawn(data, unit=self.unit)
        except TypeError:
            result = self.spawn(data, unit=self.unit)
        return result

    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Called to convert this variable to a `numpy.ndarray`.
        
        Notes
        -----
        - This method converts an instance to a `numpy.ndarray` by calling
          `numpy.asarray` on the underlying data. Therefore, if the data is
          already a `numpy.ndarray`, calling `numpy.asarray` on this instance
          will return the original array.
        """
        return numpy.asarray(self.data, *args, **kwargs)

    @property
    def size(self) -> int:
        """The number of numerical values in this instance."""
        if self._size is None:
            self._size = container.size(self.data)
        return self._size


